"""
Tests testing the publicly exposed test helper API.

Most of the testing module is tested via use in our own test suite, but
considering it's publicly-exposed code, we should have some dedicated tests to
it!

And thank goodness we're not using vanilla unittest/pytest or the word "test"
would be in here more than you can shake a buffalo at (see:
https://en.wikipedia.org/wiki/Buffalo_buffalo_Buffalo_buffalo_buffalo_buffalo_Buffalo_buffalo)
"""

from unittest.mock import Mock, patch

from fabric import Connection
from fabric.testing.base import MockRemote
from pytest import raises, fixture


@fixture(autouse=True)
def no_stdin():
    # Just neuter primary stdin read (making it return real bytes tho)
    with patch("fabric.runners.Remote.read_our_stdin", return_value=b""):
        yield


class MockRemote_:
    class mocks_by_default:
        @patch("paramiko.transport.socket")
        def prevents_real_ssh_connectivity_via_paramiko_guts(self, socket):
            with MockRemote():
                cxn = Connection(host="host")
                cxn.run("nope")
                # High level mock...
                assert isinstance(cxn.client, Mock)
                # ...prevented low level connectivity (would have been at least
                # one socket.socket() and one .connect() on result of same)
                assert not socket.mock_calls

        def does_not_run_safety_checks_when_nothing_really_expected(self):
            with MockRemote():
                cxn = Connection(host="host")
                assert isinstance(cxn.client, Mock)
                # NOTE: no run() or etc!
            # Would explode with old behavior due to always asserting transport
            # and connect method calls.

    class contextmanager_behavior:
        def calls_safety_and_stop_on_exit_with_try_finally(self):
            mr = MockRemote()
            # Stop now, before we overwrite, lest we leak the automatic mocking
            # from init time.
            mr.stop()
            mr.stop = Mock()
            mr.safety = Mock(side_effect=Exception("onoz"))
            with raises(Exception, match="onoz"):
                with mr:
                    pass
            # assert exit behavior
            mr.safety.assert_called_once_with()
            mr.stop.assert_called_once_with()

    class enable_sftp:
        def does_not_break_ssh_mocking(self):
            with MockRemote(enable_sftp=True) as mr:
                mr.expect(cmd="whoami")
                cxn = Connection(host="whatevs")
                cxn.run("whoami")
                # Safety: can call sftp stuff w/o expect()ing it, should noop
                # instead of exploding
                cxn.put("whatevs")

        def enables_per_session_sftp_mocks(self):
            with MockRemote(enable_sftp=True) as mr:
                mr.expect(
                    cmd="rm file",
                    transfers=[
                        dict(
                            method="put",
                            localpath="/local/whatevs",
                            remotepath="/remote/whatevs",
                        )
                    ],
                )
                cxn = Connection(host="host")
                cxn.run("rm file")
                cxn.put("whatevs")

        def safety_checks_work(self):
            with raises(AssertionError, match=r"put(.*whatevs)"):
                with MockRemote(enable_sftp=True) as mr:
                    mr.expect(
                        transfers=[
                            dict(
                                method="put",
                                localpath="/local/whatevs",
                                remotepath="/remote/whatevs",
                            )
                        ],
                    )
                    cxn = Connection(host="host")
                    # Satisfy the less rigorous default expectations for
                    # commands
                    cxn.run("rm file")
                    # Oh no! The wrong put()!
                    cxn.put("onoz")



"""
Tunnel and connection forwarding internals.

If you're looking for simple, end-user-focused connection forwarding, please
see `.Connection`, e.g. `.Connection.forward_local`.
"""

import select
import socket
import time
from threading import Event

from invoke.exceptions import ThreadException
from invoke.util import ExceptionHandlingThread


class TunnelManager(ExceptionHandlingThread):
    """
    Thread subclass for tunnelling connections over SSH between two endpoints.

    Specifically, one instance of this class is sufficient to sit around
    forwarding any number of individual connections made to one end of the
    tunnel or the other. If you need to forward connections between more than
    one set of ports, you'll end up instantiating multiple TunnelManagers.

    Wraps a `~paramiko.transport.Transport`, which should already be connected
    to the remote server.

    .. versionadded:: 2.0
    """

    def __init__(
        self,
        local_host,
        local_port,
        remote_host,
        remote_port,
        transport,
        finished,
    ):
        super().__init__()
        self.local_address = (local_host, local_port)
        self.remote_address = (remote_host, remote_port)
        self.transport = transport
        self.finished = finished

    def _run(self):
        # Track each tunnel that gets opened during our lifetime
        tunnels = []

        # Set up OS-level listener socket on forwarded port
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # TODO: why do we want REUSEADDR exactly? and is it portable?
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # NOTE: choosing to deal with nonblocking semantics and a fast loop,
        # versus an older approach which blocks & expects outer scope to cause
        # a socket exception by close()ing the socket.
        sock.setblocking(0)
        sock.bind(self.local_address)
        sock.listen(1)

        while not self.finished.is_set():
            # Main loop-wait: accept connections on the local listener
            # NOTE: EAGAIN means "you're nonblocking and nobody happened to
            # connect at this point in time"
            try:
                tun_sock, local_addr = sock.accept()
                # Set TCP_NODELAY to match OpenSSH's forwarding socket behavior
                tun_sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            except BlockingIOError:  # ie errno.EAGAIN
                # TODO: make configurable
                time.sleep(0.01)
                continue

            # Set up direct-tcpip channel on server end
            # TODO: refactor w/ what's used for gateways
            channel = self.transport.open_channel(
                "direct-tcpip", self.remote_address, local_addr
            )

            # Set up 'worker' thread for this specific connection to our
            # tunnel, plus its dedicated signal event (which will appear as a
            # public attr, no need to track both independently).
            finished = Event()
            tunnel = Tunnel(channel=channel, sock=tun_sock, finished=finished)
            tunnel.start()
            tunnels.append(tunnel)

        exceptions = []
        # Propogate shutdown signal to all tunnels & wait for closure
        # TODO: would be nice to have some output or at least logging here,
        # especially for "sets up a handful of tunnels" use cases like
        # forwarding nontrivial HTTP traffic.
        for tunnel in tunnels:
            tunnel.finished.set()
            tunnel.join()
            wrapper = tunnel.exception()
            if wrapper:
                exceptions.append(wrapper)
        # Handle exceptions
        if exceptions:
            raise ThreadException(exceptions)

        # All we have left to close is our own sock.
        # TODO: use try/finally?
        sock.close()


class Tunnel(ExceptionHandlingThread):
    """
    Bidirectionally forward data between an SSH channel and local socket.

    .. versionadded:: 2.0
    """

    def __init__(self, channel, sock, finished):
        self.channel = channel
        self.sock = sock
        self.finished = finished
        self.socket_chunk_size = 1024
        self.channel_chunk_size = 1024
        super().__init__()

    def _run(self):
        try:
            empty_sock, empty_chan = None, None
            while not self.finished.is_set():
                r, w, x = select.select([self.sock, self.channel], [], [], 1)
                if self.sock in r:
                    empty_sock = self.read_and_write(
                        self.sock, self.channel, self.socket_chunk_size
                    )
                if self.channel in r:
                    empty_chan = self.read_and_write(
                        self.channel, self.sock, self.channel_chunk_size
                    )
                if empty_sock or empty_chan:
                    break
        finally:
            self.channel.close()
            self.sock.close()

    def read_and_write(self, reader, writer, chunk_size):
        """
        Read ``chunk_size`` from ``reader``, writing result to ``writer``.

        Returns ``None`` if successful, or ``True`` if the read was empty.

        .. versionadded:: 2.0
        """
        data = reader.recv(chunk_size)
        if len(data) == 0:
            return True
        writer.sendall(data)



from invoke import Collection, Context, Call, Task as InvokeTask
from invoke.parser import ParseResult, ParserContext, Argument
from fabric import Executor, Task, Connection
from fabric.executor import ConnectionCall
from fabric.exceptions import NothingToDo

from unittest.mock import Mock
from pytest import skip, raises  # noqa


def _get_executor(hosts_flag=None, hosts_kwarg=None, post=None, remainder=""):
    post_tasks = []
    if post is not None:
        post_tasks.append(post)
    hosts = Argument(name="hosts")
    if hosts_flag is not None:
        hosts.value = hosts_flag
    core_args = ParseResult([ParserContext(args=[hosts])])
    core_args.remainder = remainder
    body = Mock(pre=[], post=[])
    task = Task(body, post=post_tasks, hosts=hosts_kwarg)
    coll = Collection(mytask=task)
    return body, Executor(coll, core=core_args)


def _execute(**kwargs):
    invocation = kwargs.pop("invocation", ["mytask"])
    task, executor = _get_executor(**kwargs)
    executor.execute(*invocation)
    return task


class Executor_:
    class expand_calls:
        class hosts_flag_empty:
            def no_parameterization_is_done(self):
                task = _execute()
                assert task.call_count == 1
                assert isinstance(task.call_args[0][0], Context)

        class hosts_flag_set:
            def parameterization_per_host(self):
                task = _execute(hosts_flag="host1,host2,host3")
                assert task.call_count == 3
                assert isinstance(task.call_args[0][0], Connection)

            def post_tasks_happen_once_only(self):
                post = Mock()
                task = _execute(
                    hosts_flag="host1,host2,host3", post=Task(post)
                )
                assert task.call_count == 3
                assert post.call_count == 1

        class hosts_attribute_on_task_objects:
            def parameterization_per_host(self):
                task = _execute(hosts_kwarg=["host1", "host2", "host3"])
                assert task.call_count == 3
                assert isinstance(task.call_args[0][0], Connection)

            def post_tasks_happen_once_only(self):
                post = Mock()
                task = _execute(
                    hosts_kwarg=["host1", "host2", "host3"], post=Task(post)
                )
                assert task.call_count == 3
                assert post.call_count == 1

            def may_give_Connection_kwargs_as_values(self):
                task = _execute(
                    hosts_kwarg=[
                        {"host": "host1"},
                        {"host": "host2", "user": "doge"},
                    ]
                )
                assert task.call_count == 2
                expected = [
                    Connection("host1"),
                    Connection("host2", user="doge"),
                ]
                assert [x[0][0] for x in task.call_args_list] == expected

        class Invoke_task_objects_without_hosts_attribute_still_work:
            def execution_happens_normally_without_parameterization(self):
                body = Mock(pre=[], post=[])
                coll = Collection(mytask=InvokeTask(body))
                hosts = Argument(name="hosts")
                core_args = ParseResult([ParserContext(args=[hosts])])
                # When #1824 present, this just blows up because no .hosts attr
                Executor(coll, core=core_args).execute("mytask")
                assert body.call_count == 1

            def hosts_flag_still_triggers_parameterization(self):
                body = Mock(pre=[], post=[])
                coll = Collection(mytask=InvokeTask(body))
                hosts = Argument(name="hosts")
                hosts.value = "host1,host2,host3"
                core_args = ParseResult([ParserContext(args=[hosts])])
                Executor(coll, core=core_args).execute("mytask")
                assert body.call_count == 3

        class hosts_flag_vs_attributes:
            def flag_wins(self):
                task = _execute(
                    hosts_flag="via-flag", hosts_kwarg=["via-kwarg"]
                )
                assert task.call_count == 1
                assert task.call_args[0][0] == Connection(host="via-flag")

        class remainder:
            def raises_NothingToDo_when_no_hosts(self):
                with raises(NothingToDo):
                    _execute(remainder="whatever")

            def creates_anonymous_call_per_host(self):
                # TODO: annoying to do w/o mucking around w/ our Executor class
                # more, and that stuff wants to change semi soon anyways when
                # we grow past --hosts; punting.
                skip()

        class dedupe:
            def deduplication_not_performed(self):
                task = _execute(invocation=["mytask", "mytask"])
                assert task.call_count == 2  # not 1

        class parameterize:
            def always_generates_ConnectionCall_with_host_attr(self):
                task, executor = _get_executor(hosts_flag="host1,host2,host3")
                calls = executor.expand_calls(calls=[Call(task)])
                assert len(calls) == 3
                assert all(isinstance(x, ConnectionCall) for x in calls)
                assert [x.init_kwargs["host"] for x in calls] == [
                    "host1",
                    "host2",
                    "host3",
                ]



import os
from os.path import join, dirname, abspath
from datetime import datetime

from invocations.environment import in_ci
import alabaster


# Alabaster theme + mini-extension
html_theme_path = [alabaster.get_path()]
extensions = ["alabaster", "sphinx.ext.intersphinx"]

# Paths relative to invoking conf.py - not this shared file
html_static_path = [join("..", "_shared_static")]
html_theme = "alabaster"
html_theme_options = {
    "logo": "logo.png",
    "logo_name": True,
    "logo_text_align": "center",
    "description": "Pythonic remote execution",
    "github_user": "fabric",
    "github_repo": "fabric",
    "travis_button": False,  # Circle now
    "codecov_button": False,  # README badge now
    "tidelift_url": "https://tidelift.com/subscription/pkg/pypi-fabric?utm_source=pypi-fabric&utm_medium=referral&utm_campaign=docs",
    "analytics_id": "UA-18486793-1",
    "link": "#3782BE",
    "link_hover": "#3782BE",
    # Wide enough that 80-col code snippets aren't truncated on default font
    # settings (at least for bitprophet's Chrome-on-OSX-Yosemite setup)
    "page_width": "1024px",
}
html_sidebars = {
    "**": ["about.html", "navigation.html", "searchbox.html", "donate.html"]
}

# Enable & configure doctest
extensions.append("sphinx.ext.doctest")
doctest_global_setup = r"""
from fabric.testing.base import MockRemote, MockSFTP, Session, Command
"""

on_rtd = os.environ.get("READTHEDOCS") == "True"
on_dev = not (on_rtd or in_ci())

# Invoke (docs + www)
inv_target = join(
    dirname(__file__), "..", "..", "invoke", "sites", "docs", "_build"
)
if not on_dev:
    inv_target = "https://docs.pyinvoke.org/en/latest/"
inv_www_target = join(
    dirname(__file__), "..", "..", "invoke", "sites", "www", "_build"
)
if not on_dev:
    inv_www_target = "https://pyinvoke.org/"
# Paramiko (docs)
para_target = join(
    dirname(__file__), "..", "..", "paramiko", "sites", "docs", "_build"
)
if not on_dev:
    para_target = "https://docs.paramiko.org/en/latest/"
intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "invoke": (inv_target, None),
    "invoke_www": (inv_www_target, None),
    "paramiko": (para_target, None),
}

# Regular settings
project = "Fabric"
copyright = f"{datetime.now().year} Jeff Forcier"
master_doc = "index"
templates_path = ["_templates"]
exclude_trees = ["_build"]
source_suffix = ".rst"
default_role = "obj"



# Obtain shared config values
import sys
from os.path import abspath, join, dirname

sys.path.append(abspath(join(dirname(__file__), "..")))
sys.path.append(abspath(join(dirname(__file__), "..", "..")))
from shared_conf import *

# Enable & configure autodoc
extensions.append("sphinx.ext.autodoc")
# Autodoc settings
autodoc_default_options = {
    "members": True,
    "special-members": True,
}
# TODO: consider documenting things like Remote.run usefully and
# re-enabling this? new as of sphinx 1.7 and NOT the old behavior, so very
# surprising in a bunch of spots right now, where it pulls in Invoke
# docstrings that then have bad refs and so on.
autodoc_inherit_docstrings = False

# Default is 'local' building, but reference the public WWW site when building
# under RTD.
target = join(dirname(__file__), "..", "www", "_build")
if on_rtd:
    target = "https://www.fabfile.org/"
www = (target, None)
# Intersphinx connection to www site
intersphinx_mapping.update({"www": www})

# Sister-site links to WWW
html_theme_options["extra_nav_links"] = {
    "Main website": "https://www.fabfile.org"
}



# flake8: noqa
from os.path import isfile, expanduser
from unittest.mock import patch

from pytest import fixture

# Set up icecream globally for convenience.
from icecream import install as install_icecream

from fabric.testing.fixtures import client, remote, sftp, sftp_objs, transfer


install_icecream()


# TODO: does this want to end up in the public fixtures module too?
@fixture(autouse=True)
def no_user_ssh_config():
    """
    Cowardly refuse to ever load what looks like user SSH config paths.

    Prevents the invoking user's real config from gumming up test results or
    inflating test runtime (eg if it sets canonicalization on, which will incur
    DNS lookups for nearly all of this suite's bogus names).
    """
    # An ugly, but effective, hack. I am not proud. I also don't see anything
    # that's >= as bulletproof and less ugly?
    # TODO: ideally this should expand to cover system config paths too, but
    # that's even less likely to be an issue.
    def no_config_for_you(path):
        if path == expanduser("~/.ssh/config"):
            return False
        return isfile(path)

    with patch("fabric.config.os.path.isfile", no_config_for_you):
        yield



from contextlib import contextmanager
import os
import re
import sys

from invoke.vendor.lexicon import Lexicon
from pytest_relaxed import trap

from fabric.main import make_program


support = os.path.join(os.path.abspath(os.path.dirname(__file__)), "_support")
config_file = os.path.abspath(os.path.join(support, "config.yml"))

# TODO: move invoke's support_path + load + etc somewhere importable? or into
# pytest-relaxed, despite it not being strictly related to that feature set?
# ugh
@contextmanager
def support_path():
    sys.path.insert(0, support)
    try:
        yield
    finally:
        sys.path.pop(0)


def load(name):
    with support_path():
        imported = __import__(name)
        return imported


# TODO: this could become a fixture in conftest.py, presumably, and just yield
# stdout, allowing the tests themselves to assert more naturally
@trap
def expect(invocation, out, program=None, test="equals"):
    if program is None:
        program = make_program()
    program.run("fab {}".format(invocation), exit=False)
    output = sys.stdout.getvalue()
    if test == "equals":
        assert output == out
    elif test == "contains":
        assert out in output
    elif test == "regex":
        assert re.match(out, output)
    else:
        err = "Don't know how to expect that <stdout> {} <expected>!"
        assert False, err.format(test)
    # Safety check: no stderr
    assert not sys.stderr.getvalue()


def faux_v1_env():
    # Close enough to v1 _AttributeDict...
    # Contains a copy of enough of v1's defaults to prevent us having to do a
    # lot of extra .get()s...meh
    return Lexicon(
        always_use_pty=True,
        forward_agent=False,
        gateway=None,
        host_string="localghost",
        key_filename=None,
        no_agent=False,
        password=None,
        port=22,
        ssh_config_path=None,
        # Used in a handful of safety tests, so it gets a 'real' value. eh.
        sudo_password="nope",
        sudo_prompt=None,
        timeout=None,
        use_ssh_config=False,
        user="localuser",
        warn_only=False,
    )



from fabric import task


@task
def expect_conf_value(c):
    assert c.it_came_from == "yml"


@task
def expect_conf_key_filename(c):
    expected = ["private.key", "other.key"]
    got = c.connect_kwargs.key_filename
    assert got == expected, "{!r} != {!r}".format(got, expected)


@task
def expect_cli_key_filename(c):
    expected = ["cli.key"]
    got = c.connect_kwargs.key_filename
    assert got == expected, "{!r} != {!r}".format(got, expected)



from functools import partial
from getpass import getpass
from pathlib import Path

from paramiko import Agent, PKey
from paramiko.auth_strategy import (
    AuthStrategy,
    Password,
    InMemoryPrivateKey,
    OnDiskPrivateKey,
)

from .util import win32


class OpenSSHAuthStrategy(AuthStrategy):
    """
    Auth strategy that tries very hard to act like the OpenSSH client.

    .. warning::
        As of version 3.1, this class is **EXPERIMENTAL** and **incomplete**.
        It works best with passphraseless (eg ssh-agent) private key auth for
        now and will grow more features in future releases.

    For example, it accepts a `~paramiko.config.SSHConfig` and uses any
    relevant ``IdentityFile`` directives from that object, along with keys from
    your home directory and any local SSH agent. Keys specified at runtime are
    tried last, just as with ``ssh -i /path/to/key`` (this is one departure
    from the legacy/off-spec auth behavior observed in older Paramiko and
    Fabric versions).

    We explicitly do not document the full details here, because the point is
    to match the documented/observed behavior of OpenSSH. Please see the `ssh
    <https://man.openbsd.org/ssh>`_ and `ssh_config
    <https://man.openbsd.org/ssh_config>`_ man pages for more information.

    .. versionadded:: 3.1
    """

    # Skimming openssh code (ssh.c and sshconnect2.c) gives us the following
    # behavior to crib from:
    # - parse cli (initializing identity_files if any given)
    # - parse user config, then system config _unless_ user gave cli config
    # path; this will also add to identity_files if any IdentityFile found
    # (after the CLI ones)
    # - lots of value init, string interpolation, etc
    # - if no other identity_files exist at this point, fill in the ~/.ssh/
    # defaults:
    #   - in order: rsa, dsa, ecdsa, ecdsa_sk, ed25519, xmss (???)
    # - initial connection (ssh_connect() - presumably handshake/hostkey/kex)
    # - load all identity_files (and any implicit certs of those)
    # - eventually runs pubkey_prepare() which, per its own comment,
    # loads/assembles key material in this order:
    #   - certs - config file, then cli, skipping any non-user (?) certs
    #   - agent keys that are also listed in the config file
    #   - agent keys _not_ listed in config files
    #   - non-agent config file keys (this seems like it includes cli and
    #   implicit defaults)
    #   - once list is assembled, drop anything not listed in configured pubkey
    #   algorithms list
    # - auth_none to get list of acceptable authmethods
    # - while-loops along that, or next returned, list of acceptable
    # authmethods, using a handler table, so eg a 'standard' openssh on both
    # ends might start with 'publickey,password,keyboard-interactive'; so it'll
    # try all pubkeys found above before eventually trying a password prompt,
    # and then if THAT fails, it will try kbdint call-and-response (similar to
    # password but where server sends you the prompt(s) it wants displayed)

    def __init__(self, ssh_config, fabric_config, username):
        """
        Extends superclass with additional inputs.

        Specifically:

        - ``fabric_config``, a `fabric.Config` instance for the current
          session.
        - ``username``, which is unified by our intended caller so we don't
          have to - it's a synthesis of CLI, runtime,
          invoke/fabric-configuration, and ssh_config configuration.

        Also handles connecting to an SSH agent, if possible, for easier
        lifecycle tracking.
        """
        super().__init__(ssh_config=ssh_config)
        self.username = username
        self.config = fabric_config
        # NOTE: Agent seems designed to always 'work' even w/o a live agent, in
        # which case it just yields an empty key list.
        self.agent = Agent()

    def get_pubkeys(self):
        # Similar to OpenSSH, we first obtain sources in arbitrary order,
        # tracking where they came from and whether they were a cert.
        config_certs, config_keys, cli_certs, cli_keys = [], [], [], []
        # Our own configuration is treated like `ssh -i`, partly because that's
        # where our -i flag ends up, partly because our config data has no
        # direct OpenSSH analogue (it's _not_ in your ssh_config! it's
        # elsewhere!)
        for path in self.config.authentication.identities:
            try:
                key = PKey.from_path(path)
            except FileNotFoundError:
                continue
            source = OnDiskPrivateKey(
                username=self.username,
                source="python-config",
                path=path,
                pkey=key,
            )
            (cli_certs if key.public_blob else cli_keys).append(source)
        # Now load ssh_config IdentityFile directives, sorting again into
        # cert/key split.
        # NOTE: Config's ssh_config loader already behaves OpenSSH-compatibly:
        # if the user supplied a custom ssh_config file path, that is the only
        # one loaded; otherwise, it loads and merges the user and system paths.
        # TODO: CertificateFile support? Most people seem to rely on the
        # implicit cert loading of IdentityFile...
        for path in self.ssh_config.get("identityfile", []):
            try:
                key = PKey.from_path(path)
            except FileNotFoundError:
                continue
            source = OnDiskPrivateKey(
                username=self.username,
                source="ssh-config",
                path=path,
                pkey=key,
            )
            (config_certs if key.public_blob else config_keys).append(source)
        # At this point, if we've still got no keys or certs, look in the
        # default user locations.
        if not any((config_certs, config_keys, cli_certs, cli_keys)):
            user_ssh = Path.home() / f"{'' if win32 else '.'}ssh"
            # This is the same order OpenSSH documents as using.
            for type_ in ("rsa", "ecdsa", "ed25519", "dsa"):
                path = user_ssh / f"id_{type_}"
                try:
                    key = PKey.from_path(path)
                except FileNotFoundError:
                    continue
                source = OnDiskPrivateKey(
                    username=self.username,
                    source="implicit-home",
                    path=path,
                    pkey=key,
                )
                dest = config_certs if key.public_blob else config_keys
                dest.append(source)
        # TODO: set agent_keys to empty list if IdentitiesOnly is true
        agent_keys = self.agent.get_keys()

        # We've finally loaded everything; now it's time to throw them upwards
        # in the intended order...
        # TODO: define subroutine that dedupes (& honors
        # PubkeyAcceptedAlgorithms) then rub that on all of the below.
        # First, all local _certs_ (config wins over cli, for w/e reason)
        for source in config_certs:
            yield source
        for source in cli_certs:
            yield source
        # Then all agent keys, first ones that were also mentioned in configs,
        # then 'new' ones not found in configs.
        deferred_agent_keys = []
        for key in agent_keys:
            config_index = None
            for i, config_key in enumerate(config_keys):
                if config_key.pkey == key:
                    config_index = i
                    break
            if config_index:
                yield InMemoryPrivateKey(username=self.username, pkey=key)
                # Nuke so it doesn't get re-yielded by regular conf keys bit
                del config_keys[config_index]
            else:
                deferred_agent_keys.append(key)
        for key in deferred_agent_keys:
            yield InMemoryPrivateKey(username=self.username, pkey=key)
        for source in cli_keys:
            yield source
        # This will now be just the config-borne keys that were NOT in agent
        for source in config_keys:
            yield source

    def get_sources(self):
        # TODO: initial none-auth + tracking the response's allowed types.
        # however, SSHClient never did this deeply, and there's no guarantee a
        # server _will_ send anything but "any" anyways...
        # Public keys of all kinds typically first.
        yield from self.get_pubkeys()
        user = self.username
        prompter = partial(getpass, f"{user}'s password: ")
        # Then password.
        yield Password(username=self.username, password_getter=prompter)

    def authenticate(self, *args, **kwargs):
        # Just do what our parent would, except make sure we close() after.
        try:
            return super().authenticate(*args, **kwargs)
        finally:
            self.close()

    def close(self):
        """
        Shut down any resources we ourselves opened up.
        """
        # TODO: bare try/except here as "best effort"? ugh
        self.agent.close()



"""
File transfer via SFTP and/or SCP.
"""

import os
import posixpath
import stat

from pathlib import Path

from .util import debug  # TODO: actual logging! LOL

# TODO: figure out best way to direct folks seeking rsync, to patchwork's rsync
# call (which needs updating to use invoke.run() & fab 2 connection methods,
# but is otherwise suitable).
# UNLESS we want to try and shoehorn it into this module after all? Delegate
# any recursive get/put to it? Requires users to have rsync available of
# course.


class Transfer:
    """
    `.Connection`-wrapping class responsible for managing file upload/download.

    .. versionadded:: 2.0
    """

    # TODO: SFTP clear default, but how to do SCP? subclass? init kwarg?

    def __init__(self, connection):
        self.connection = connection

    @property
    def sftp(self):
        return self.connection.sftp()

    def is_remote_dir(self, path):
        try:
            return stat.S_ISDIR(self.sftp.stat(path).st_mode)
        except IOError:
            return False

    def get(self, remote, local=None, preserve_mode=True):
        """
        Copy a file from wrapped connection's host to the local filesystem.

        :param str remote:
            Remote file to download.

            May be absolute, or relative to the remote working directory.

            .. note::
                Most SFTP servers set the remote working directory to the
                connecting user's home directory, and (unlike most shells) do
                *not* expand tildes (``~``).

                For example, instead of saying ``get("~/tmp/archive.tgz")``,
                say ``get("tmp/archive.tgz")``.

        :param local:
            Local path to store downloaded file in, or a file-like object.

            **If None or another 'falsey'/empty value is given** (the default),
            the remote file is downloaded to the current working directory (as
            seen by `os.getcwd`) using its remote filename. (This is equivalent
            to giving ``"{basename}"``; see the below subsection on
            interpolation.)

            **If a string is given**, it should be a path to a local directory
            or file and is subject to similar behavior as that seen by common
            Unix utilities or OpenSSH's ``sftp`` or ``scp`` tools.

            For example, if the local path is a directory, the remote path's
            base filename will be added onto it (so ``get('foo/bar/file.txt',
            '/tmp/')`` would result in creation or overwriting of
            ``/tmp/file.txt``).

            This path will be **interpolated** with some useful parameters,
            using `str.format`:

            - The `.Connection` object's ``host``, ``user`` and ``port``
              attributes.
            - The ``basename`` and ``dirname`` of the ``remote`` path, as
              derived by `os.path` (specifically, its ``posixpath`` flavor, so
              that the resulting values are useful on remote POSIX-compatible
              SFTP servers even if the local client is Windows).
            - Thus, for example, ``"/some/path/{user}@{host}/{basename}"`` will
              yield different local paths depending on the properties of both
              the connection and the remote path.

            .. note::
                If nonexistent directories are present in this path (including
                the final path component, if it ends in `os.sep`) they will be
                created automatically using `os.makedirs`.

            **If a file-like object is given**, the contents of the remote file
            are simply written into it.

        :param bool preserve_mode:
            Whether to `os.chmod` the local file so it matches the remote
            file's mode (default: ``True``).

        :returns: A `.Result` object.

        .. versionadded:: 2.0
        .. versionchanged:: 2.6
            Added ``local`` path interpolation of connection & remote file
            attributes.
        .. versionchanged:: 2.6
            Create missing ``local`` directories automatically.
        """
        # TODO: how does this API change if we want to implement
        # remote-to-remote file transfer? (Is that even realistic?)
        # TODO: callback support
        # TODO: how best to allow changing the behavior/semantics of
        # remote/local (e.g. users might want 'safer' behavior that complains
        # instead of overwriting existing files) - this likely ties into the
        # "how to handle recursive/rsync" and "how to handle scp" questions

        # Massage remote path
        if not remote:
            raise ValueError("Remote path must not be empty!")
        orig_remote = remote
        remote = posixpath.join(
            self.sftp.getcwd() or self.sftp.normalize("."), remote
        )

        # Massage local path
        orig_local = local
        is_file_like = hasattr(local, "write") and callable(local.write)
        remote_filename = posixpath.basename(remote)
        if not local:
            local = remote_filename
        # Path-driven local downloads need interpolation, abspath'ing &
        # directory creation
        if not is_file_like:
            local = local.format(
                host=self.connection.host,
                user=self.connection.user,
                port=self.connection.port,
                dirname=posixpath.dirname(remote),
                basename=remote_filename,
            )
            # Must treat dir vs file paths differently, lest we erroneously
            # mkdir what was intended as a filename, and so that non-empty
            # dir-like paths still get remote filename tacked on.
            if local.endswith(os.sep):
                dir_path = local
                local = os.path.join(local, remote_filename)
            else:
                dir_path, _ = os.path.split(local)
            local = os.path.abspath(local)
            Path(dir_path).mkdir(parents=True, exist_ok=True)
            # TODO: reimplement mkdir (or otherwise write a testing function)
            # allowing us to track what was created so we can revert if
            # transfer fails.
            # TODO: Alternately, transfer to temp location and then move, but
            # that's basically inverse of v1's sudo-put which gets messy

        # Run Paramiko-level .get() (side-effects only. womp.)
        # TODO: push some of the path handling into Paramiko; it should be
        # responsible for dealing with path cleaning etc.
        # TODO: probably preserve warning message from v1 when overwriting
        # existing files. Use logging for that obviously.
        #
        # If local appears to be a file-like object, use sftp.getfo, not get
        if is_file_like:
            self.sftp.getfo(remotepath=remote, fl=local)
        else:
            self.sftp.get(remotepath=remote, localpath=local)
            # Set mode to same as remote end
            # TODO: Push this down into SFTPClient sometime (requires backwards
            # incompat release.)
            if preserve_mode:
                remote_mode = self.sftp.stat(remote).st_mode
                mode = stat.S_IMODE(remote_mode)
                os.chmod(local, mode)
        # Return something useful
        return Result(
            orig_remote=orig_remote,
            remote=remote,
            orig_local=orig_local,
            local=local,
            connection=self.connection,
        )

    def put(self, local, remote=None, preserve_mode=True):
        """
        Upload a file from the local filesystem to the current connection.

        :param local:
            Local path of file to upload, or a file-like object.

            **If a string is given**, it should be a path to a local (regular)
            file (not a directory).

            .. note::
                When dealing with nonexistent file paths, normal Python file
                handling concerns come into play - for example, trying to
                upload a nonexistent ``local`` path will typically result in an
                `OSError`.

            **If a file-like object is given**, its contents are written to the
            remote file path.

        :param str remote:
            Remote path to which the local file will be written.

            .. note::
                Most SFTP servers set the remote working directory to the
                connecting user's home directory, and (unlike most shells) do
                *not* expand tildes (``~``).

                For example, instead of saying ``put("archive.tgz",
                "~/tmp/")``, say ``put("archive.tgz", "tmp/")``.

                In addition, this means that 'falsey'/empty values (such as the
                default value, ``None``) are allowed and result in uploading to
                the remote home directory.

            .. note::
                When ``local`` is a file-like object, ``remote`` is required
                and must refer to a valid file path (not a directory).

        :param bool preserve_mode:
            Whether to ``chmod`` the remote file so it matches the local file's
            mode (default: ``True``).

        :returns: A `.Result` object.

        .. versionadded:: 2.0
        """
        if not local:
            raise ValueError("Local path must not be empty!")

        is_file_like = hasattr(local, "write") and callable(local.write)

        # Massage remote path
        orig_remote = remote
        if is_file_like:
            local_base = getattr(local, "name", None)
        else:
            local_base = os.path.basename(local)
        if not remote:
            if is_file_like:
                raise ValueError(
                    "Must give non-empty remote path when local is a file-like object!"  # noqa
                )
            else:
                remote = local_base
                debug("Massaged empty remote path into {!r}".format(remote))
        elif self.is_remote_dir(remote):
            # non-empty local_base implies a) text file path or b) FLO which
            # had a non-empty .name attribute. huzzah!
            if local_base:
                remote = posixpath.join(remote, local_base)
            else:
                if is_file_like:
                    raise ValueError(
                        "Can't put a file-like-object into a directory unless it has a non-empty .name attribute!"  # noqa
                    )
                else:
                    # TODO: can we ever really end up here? implies we want to
                    # reorganize all this logic so it has fewer potential holes
                    raise ValueError(
                        "Somehow got an empty local file basename ({!r}) when uploading to a directory ({!r})!".format(  # noqa
                            local_base, remote
                        )
                    )

        prejoined_remote = remote
        remote = posixpath.join(
            self.sftp.getcwd() or self.sftp.normalize("."), remote
        )
        if remote != prejoined_remote:
            msg = "Massaged relative remote path {!r} into {!r}"
            debug(msg.format(prejoined_remote, remote))

        # Massage local path
        orig_local = local
        if not is_file_like:
            local = os.path.abspath(local)
            if local != orig_local:
                debug(
                    "Massaged relative local path {!r} into {!r}".format(
                        orig_local, local
                    )
                )  # noqa

        # Run Paramiko-level .put() (side-effects only. womp.)
        # TODO: push some of the path handling into Paramiko; it should be
        # responsible for dealing with path cleaning etc.
        # TODO: probably preserve warning message from v1 when overwriting
        # existing files. Use logging for that obviously.
        #
        # If local appears to be a file-like object, use sftp.putfo, not put
        if is_file_like:
            msg = "Uploading file-like object {!r} to {!r}"
            debug(msg.format(local, remote))
            pointer = local.tell()
            try:
                local.seek(0)
                self.sftp.putfo(fl=local, remotepath=remote)
            finally:
                local.seek(pointer)
        else:
            debug("Uploading {!r} to {!r}".format(local, remote))
            self.sftp.put(localpath=local, remotepath=remote)
            # Set mode to same as local end
            # TODO: Push this down into SFTPClient sometime (requires backwards
            # incompat release.)
            if preserve_mode:
                local_mode = os.stat(local).st_mode
                mode = stat.S_IMODE(local_mode)
                self.sftp.chmod(remote, mode)
        # Return something useful
        return Result(
            orig_remote=orig_remote,
            remote=remote,
            orig_local=orig_local,
            local=local,
            connection=self.connection,
        )


class Result:
    """
    A container for information about the result of a file transfer.

    See individual attribute/method documentation below for details.

    .. note::
        Unlike similar classes such as `invoke.runners.Result` or
        `fabric.runners.Result` (which have a concept of "warn and return
        anyways on failure") this class has no useful truthiness behavior. If a
        file transfer fails, some exception will be raised, either an `OSError`
        or an error from within Paramiko.

    .. versionadded:: 2.0
    """

    # TODO: how does this differ from put vs get? field stating which? (feels
    # meh) distinct classes differing, for now, solely by name? (also meh)
    def __init__(self, local, orig_local, remote, orig_remote, connection):
        #: The local path the file was saved as, or the object it was saved
        #: into if a file-like object was given instead.
        #:
        #: If a string path, this value is massaged to be absolute; see
        #: `.orig_local` for the original argument value.
        self.local = local
        #: The original value given as the returning method's ``local``
        #: argument.
        self.orig_local = orig_local
        #: The remote path downloaded from. Massaged to be absolute; see
        #: `.orig_remote` for the original argument value.
        self.remote = remote
        #: The original argument value given as the returning method's
        #: ``remote`` argument.
        self.orig_remote = orig_remote
        #: The `.Connection` object this result was obtained from.
        self.connection = connection

    # TODO: ensure str/repr makes it easily differentiable from run() or
    # local() result objects (and vice versa).



from invoke import Context, task as invtask
from fabric import task, Connection


@task
def build(c):
    pass


@task
def deploy(c):
    pass


@task
def basic_run(c):
    c.run("nope")


@task
def expect_vanilla_Context(c):
    assert isinstance(c, Context)
    assert not isinstance(c, Connection)


@task
def expect_from_env(c):
    assert c.config.run.echo is True


@task
def expect_mutation_to_fail(c):
    # If user level config changes are preserved between parameterized per-host
    # task calls, this would assert on subsequent invocations...
    assert "foo" not in c.config
    # ... because of this:
    c.config.foo = "bar"


@task
def mutate(c):
    c.foo = "bar"


@task
def expect_mutation(c):
    assert c.foo == "bar"


@task
def expect_identity(c):
    assert c.config.connect_kwargs["key_filename"] == ["identity.key"]


@task
def expect_identities(c):
    assert c.config.connect_kwargs["key_filename"] == [
        "identity.key",
        "identity2.key",
    ]


@task
def expect_connect_timeout(c):
    assert c.config.connect_kwargs["timeout"] == 5


@task
def first(c):
    print("First!")


@task
def third(c):
    print("Third!")


@task(pre=[first], post=[third])
def second(c, show_host=False):
    if show_host:
        print("Second: {}".format(c.host))
    else:
        print("Second!")


@task(hosts=["myhost"])
def hosts_are_myhost(c):
    c.run("nope")


@task(hosts=["host1", "host2"])
def two_hosts(c):
    c.run("nope")


@task(hosts=["someuser@host1:1234"])
def hosts_are_host_stringlike(c):
    c.run("nope")


@task(hosts=["admin@host1", {"host": "host2"}])
def hosts_are_mixed_values(c):
    c.run("nope")


@task(hosts=[{"host": "host1", "user": "admin"}, {"host": "host2"}])
def hosts_are_init_kwargs(c):
    c.run("nope")


@invtask
def vanilla_Task_works_ok(c):
    assert isinstance(c, Context)
    assert not isinstance(c, Connection)



"""
Tests concerned with the ``fab`` tool & how it overrides Invoke defaults.
"""

import os
import sys
import re

from invoke import run
from invoke.util import cd
from unittest.mock import patch
from paramiko.agent import AgentKey, Message
import pytest  # because WHY would you expose @skip normally? -_-
from pytest_relaxed import raises

from fabric.config import Config
from fabric.main import make_program
from fabric.exceptions import NothingToDo

from fabric.testing.base import Session
from _util import expect, support, config_file, trap


# Designate a runtime config file intended for the test environment; it does
# things like automatically mute stdin so test harnesses that care about stdin
# don't get upset.
# NOTE: this requires the test environment to have Invoke 1.1.0 or above; for
# now this is fine as we don't do a big serious matrix, we typically use Invoke
# master to allow testing in-dev changes.
# TODO: if that _changes_ then we may have to rethink this so that it goes back
# to being testable on Invoke >=1.0 instead of >=1.1...
os.environ["INVOKE_RUNTIME_CONFIG"] = config_file


class Fab_:
    class core_program_behavior:
        def version_output_contains_our_name_plus_deps(self):
            expect(
                "--version",
                r"""
Fabric .+
Paramiko .+
Invoke .+
""".strip(),
                test="regex",
            )

        def help_output_says_fab(self):
            expect("--help", "Usage: fab", test="contains")

        def exposes_hosts_flag_in_help(self):
            expect("--help", "-H STRING, --hosts=STRING", test="contains")

        def executes_remainder_as_anonymous_task(self, remote):
            remote.expect(host="myhost", cmd="whoami")
            make_program().run("fab -H myhost -- whoami", exit=False)

        @patch("paramiko.agent.Agent.get_keys")
        def can_list_agent_keys(self, get_keys):
            agent_keys = []
            for type_, bits, comment in (
                ("ecdsa", b"dummy", "woody"),
                ("rsa", b"ventriloquist", "bob"),
                ("ed25519", b"stagehand", "smith"),
            ):
                # Looks like a pubkey blob from an agent
                m = Message()
                m.add_string(type_)
                m.add_string(bits)
                agent_keys.append(
                    AgentKey(agent=None, blob=bytes(m), comment=comment)
                )

            get_keys.return_value = agent_keys
            expected = """
0 SHA256:r7SOU1pAlEWmRE57Swf0OQHg9tlYicKaLx2DxGbDVk8 woody (ECDSA)
0 SHA256:2qZYGN+eIVfmhwpQUMje7uG4+7tZquM5LBwNaHCBsqg bob (RSA)
0 SHA256:4seJT+aN1aTPIudGupnXsZ1z20r+GCIAAKEA4MHnwvA smith (ED25519)
""".lstrip()
            expect("--list-agent-keys", expected)

        def uses_FABRIC_env_prefix(self, environ):
            environ["FABRIC_RUN_ECHO"] = "1"
            with cd(support):
                make_program().run("fab expect-from-env")

        def basic_pre_and_post_tasks_still_work(self):
            with cd(support):
                # Safety
                expect("first", "First!\n")
                expect("third", "Third!\n")
                # Real test
                expect("second", "First!\nSecond!\nThird!\n")

    class filenames:
        def loads_fabfile_not_tasks(self):
            "Loads fabfile.py, not tasks.py"
            with cd(support):
                expect(
                    "--list",
                    """
Available tasks:

  basic-run
  build
  deploy
  expect-connect-timeout
  expect-from-env
  expect-identities
  expect-identity
  expect-mutation
  expect-mutation-to-fail
  expect-vanilla-Context
  first
  hosts-are-host-stringlike
  hosts-are-init-kwargs
  hosts-are-mixed-values
  hosts-are-myhost
  mutate
  second
  third
  two-hosts
  vanilla-Task-works-ok

""".lstrip(),
                )

        def loads_fabric_config_files_not_invoke_ones(self):
            for type_ in ("yaml", "yml", "json", "py"):
                with cd(os.path.join(support, "{}_conf".format(type_))):
                    # This task, in each subdir, expects data present in a
                    # fabric.<ext> nearby to show up in the config.
                    make_program().run("fab expect-conf-value")

    class runtime_ssh_config_path:
        def _run(
            self,
            flag="-S",
            file_="ssh_config/runtime.conf",
            tasks="runtime-ssh-config",
        ):
            with cd(support):
                # Relies on asserts within the task, which will bubble up as
                # it's executed in-process
                cmd = "fab -c runtime_fabfile {} {} -H runtime {}"
                make_program().run(cmd.format(flag, file_, tasks))

        def capital_F_flag_specifies_runtime_ssh_config_file(self):
            self._run(flag="-S")

        def long_form_flag_also_works(self):
            self._run(flag="--ssh-config")

        @raises(IOError)
        def IOErrors_if_given_missing_file(self):
            self._run(file_="nope/nothere.conf")

        @patch.object(Config, "_load_ssh_file")
        def config_only_loaded_once_per_session(self, method):
            # Task that doesn't make assertions about the config (since the
            # _actual_ config it gets is empty as we had to mock out the loader
            # method...sigh)
            self._run(tasks="dummy dummy")
            # Called only once (initial __init__) with runtime conf, instead of
            # that plus a few more pairs of calls against the default files
            # (which is what happens when clone() isn't preserving the
            # already-parsed/loaded SSHConfig)
            method.assert_called_once_with("ssh_config/runtime.conf")

    class hosts_flag_parameterizes_tasks:
        # NOTE: many of these just rely on MockRemote's builtin
        # "channel.exec_command called with given command string" asserts.

        def single_string_is_single_host_and_single_exec(self, remote):
            remote.expect(host="myhost", cmd="nope")
            # In addition to just testing a base case, this checks for a really
            # dumb bug where one appends to, instead of replacing, the task
            # list during parameterization/expansion XD
            with cd(support):
                make_program().run("fab -H myhost basic-run")

        def comma_separated_string_is_multiple_hosts(self, remote):
            remote.expect_sessions(
                Session("host1", cmd="nope"), Session("host2", cmd="nope")
            )
            with cd(support):
                make_program().run("fab -H host1,host2 basic-run")

        def multiple_hosts_works_with_remainder_too(self, remote):
            remote.expect_sessions(
                Session("host1", cmd="whoami"), Session("host2", cmd="whoami")
            )
            make_program().run("fab -H host1,host2 -- whoami")

        def host_string_shorthand_is_passed_through(self, remote):
            remote.expect(host="host1", port=1234, user="someuser")
            make_program().run("fab -H someuser@host1:1234 -- whoami")

        # NOTE: no mocking because no actual run() under test, only
        # parameterization
        # TODO: avoiding for now because implementing this requires more work
        # at the Invoke level re: deciding when to _not_ pass in the
        # session-global config object (Executor's self.config). At the moment,
        # our threading-concurrency API is oriented around Group, and we're not
        # using it for --hosts, so it's not broken...yet.
        @pytest.mark.skip
        def config_mutation_not_preserved(self):
            with cd(support):
                make_program().run(
                    "fab -H host1,host2 expect-mutation-to-fail"
                )

        @trap
        def pre_post_tasks_are_not_parameterized_across_hosts(self):
            with cd(support):
                make_program().run(
                    "fab -H hostA,hostB,hostC second --show-host"
                )
                output = sys.stdout.getvalue()
                # Expect pre once, 3x main, post once, as opposed to e.g. both
                # pre and main task
                expected = """
First!
Second: hostA
Second: hostB
Second: hostC
Third!
""".lstrip()
                assert output == expected

    class hosts_task_arg_parameterizes_tasks:
        # NOTE: many of these just rely on MockRemote's builtin
        # "channel.exec_command called with given command string" asserts.

        def single_string_is_single_exec(self, remote):
            remote.expect(host="myhost", cmd="nope")
            with cd(support):
                make_program().run("fab hosts-are-myhost")

        def multiple_strings_is_multiple_host_args(self, remote):
            remote.expect_sessions(
                Session("host1", cmd="nope"), Session("host2", cmd="nope")
            )
            with cd(support):
                make_program().run("fab two-hosts")

        def host_string_shorthand_works_ok(self, remote):
            remote.expect(host="host1", port=1234, user="someuser")
            with cd(support):
                make_program().run("fab hosts-are-host-stringlike")

        def may_give_Connection_init_kwarg_dicts(self, remote):
            remote.expect_sessions(
                Session("host1", user="admin", cmd="nope"),
                Session("host2", cmd="nope"),
            )
            with cd(support):
                make_program().run("fab hosts-are-init-kwargs")

        def may_give_mixed_value_types(self, remote):
            remote.expect_sessions(
                Session("host1", user="admin", cmd="nope"),
                Session("host2", cmd="nope"),
            )
            with cd(support):
                make_program().run("fab hosts-are-mixed-values")

    class no_hosts_flag_or_task_arg:
        def calls_task_once_with_invoke_context(self):
            with cd(support):
                make_program().run("fab expect-vanilla-Context")

        def vanilla_Invoke_task_works_too(self):
            with cd(support):
                make_program().run("fab vanilla-Task-works-ok")

        @raises(NothingToDo)
        def generates_exception_if_combined_with_remainder(self):
            make_program().run("fab -- nope")

        def invokelike_multitask_invocation_preserves_config_mutation(self):
            # Mostly a guard against Executor subclass tweaks breaking Invoke
            # behavior added in pyinvoke/invoke#309
            with cd(support):
                make_program().run("fab mutate expect-mutation")

    class connect_timeout:
        def dash_t_supplies_default_connect_timeout(self):
            with cd(support):
                make_program().run("fab -t 5 expect-connect-timeout")

        def double_dash_connect_timeout_also_works(self):
            with cd(support):
                make_program().run(
                    "fab --connect-timeout 5 expect-connect-timeout"
                )

    class runtime_identity_file:
        def dash_i_supplies_default_connect_kwarg_key_filename(self):
            # NOTE: the expect-identity task in tests/_support/fabfile.py
            # performs asserts about its context's .connect_kwargs value,
            # relying on other tests to prove connect_kwargs makes its way into
            # that context.
            with cd(support):
                make_program().run("fab -i identity.key expect-identity")

        def double_dash_identity_also_works(self):
            with cd(support):
                make_program().run(
                    "fab --identity identity.key expect-identity"
                )

        def may_be_given_multiple_times(self):
            with cd(support):
                make_program().run(
                    "fab -i identity.key -i identity2.key expect-identities"
                )

    class secrets_prompts:
        @patch("fabric.main.getpass.getpass")
        def _expect_prompt(self, getpass, flag, key, value, prompt):
            getpass.return_value = value
            with cd(support):
                # Expect that the given key was found in the context.
                cmd = "fab -c prompting {} expect-connect-kwarg --key {} --val {}"  # noqa
                make_program().run(cmd.format(flag, key, value))
            # Then we also expect that getpass was called w/ expected prompt
            getpass.assert_called_once_with(prompt)

        def password_prompt_updates_connect_kwargs(self):
            self._expect_prompt(
                flag="--prompt-for-login-password",
                key="password",
                value="mypassword",
                prompt="Enter login password for use with SSH auth: ",
            )

        def passphrase_prompt_updates_connect_kwargs(self):
            self._expect_prompt(
                flag="--prompt-for-passphrase",
                key="passphrase",
                value="mypassphrase",
                prompt="Enter passphrase for use unlocking SSH keys: ",
            )

    class configuration_updating_and_merging:
        def key_filename_can_be_set_via_non_override_config_levels(self):
            # Proves/protects against #1762, where eg key_filenames gets
            # 'reset' to an empty list. Arbitrarily uses the 'yml' level of
            # test fixtures, which has a fabric.yml w/ a
            # connect_kwargs.key_filename value of [private.key, other.key].
            with cd(os.path.join(support, "yml_conf")):
                make_program().run("fab expect-conf-key-filename")

        def cli_identity_still_overrides_when_non_empty(self):
            with cd(os.path.join(support, "yml_conf")):
                make_program().run("fab -i cli.key expect-cli-key-filename")

    class completion:
        # NOTE: most completion tests are in Invoke too; this is just an
        # irritating corner case driven by Fabric's 'remainder' functionality.
        @trap
        def complete_flag_does_not_trigger_remainder_only_behavior(self):
            # When bug present, 'fab --complete -- fab' fails to load any
            # collections because it thinks it's in remainder-only,
            # work-without-a-collection mode.
            with cd(support):
                make_program().run("fab --complete -- fab", exit=False)
            # Cherry-picked safety checks looking for tasks from fixture
            # fabfile
            output = sys.stdout.getvalue()
            for name in ("build", "deploy", "expect-from-env"):
                assert name in output


class main:
    "__main__"

    def python_dash_m_acts_like_fab(self, capsys):
        # Rehash of version output test, but using 'python -m fabric'
        expected_output = r"""
Fabric .+
Paramiko .+
Invoke .+
""".strip()
        output = run("python -m fabric --version", hide=True, in_stream=False)
        assert re.match(expected_output, output.stdout)



# flake8: noqa
from ._version import __version_info__, __version__
from .connection import Config, Connection
from .runners import Remote, RemoteShell, Result
from .group import Group, SerialGroup, ThreadingGroup, GroupResult
from .tasks import task, Task
from .executor import Executor

# Best-effort import of module relying on a Paramiko 3.2+ API member
# TODO: this is chiefly a concession to our "v1->v2 shim test" in CI, since
# Fabric 1.x wants Paramiko<3.
try:
    from .auth import OpenSSHAuthStrategy
except ImportError:
    pass



from contextlib import contextmanager
from io import StringIO
from threading import Event
import socket

from decorator import decorator
from invoke import Context
from invoke.exceptions import ThreadException
from paramiko.agent import AgentRequestHandler
from paramiko.client import SSHClient, AutoAddPolicy
from paramiko.config import SSHConfig
from paramiko.proxy import ProxyCommand

from .config import Config
from .exceptions import InvalidV1Env
from .transfer import Transfer
from .tunnels import TunnelManager, Tunnel


@decorator
def opens(method, self, *args, **kwargs):
    self.open()
    return method(self, *args, **kwargs)


def derive_shorthand(host_string):
    user_hostport = host_string.rsplit("@", 1)
    hostport = user_hostport.pop()
    user = user_hostport[0] if user_hostport and user_hostport[0] else None

    # IPv6: can't reliably tell where addr ends and port begins, so don't
    # try (and don't bother adding special syntax either, user should avoid
    # this situation by using port=).
    if hostport.count(":") > 1:
        host = hostport
        port = None
    # IPv4: can split on ':' reliably.
    else:
        host_port = hostport.rsplit(":", 1)
        host = host_port.pop(0) or None
        port = host_port[0] if host_port and host_port[0] else None

    if port is not None:
        port = int(port)

    return {"user": user, "host": host, "port": port}


class Connection(Context):
    """
    A connection to an SSH daemon, with methods for commands and file transfer.

    **Basics**

    This class inherits from Invoke's `~invoke.context.Context`, as it is a
    context within which commands, tasks etc can operate. It also encapsulates
    a Paramiko `~paramiko.client.SSHClient` instance, performing useful high
    level operations with that `~paramiko.client.SSHClient` and
    `~paramiko.channel.Channel` instances generated from it.

    .. _connect_kwargs:

    .. note::
        Many SSH specific options -- such as specifying private keys and
        passphrases, timeouts, disabling SSH agents, etc -- are handled
        directly by Paramiko and should be specified via the
        :ref:`connect_kwargs argument <connect_kwargs-arg>` of the constructor.

    **Lifecycle**

    `.Connection` has a basic "`create <__init__>`, `connect/open <open>`, `do
    work <run>`, `disconnect/close <close>`" lifecycle:

    - `Instantiation <__init__>` imprints the object with its connection
      parameters (but does **not** actually initiate the network connection).

        - An alternate constructor exists for users :ref:`upgrading piecemeal
          from Fabric 1 <from-v1>`: `from_v1`

    - Methods like `run`, `get` etc automatically trigger a call to
      `open` if the connection is not active; users may of course call `open`
      manually if desired.
    - It's best to explicitly close your connections when done using them. This
      can be accomplished by manually calling `close`, or by using the object
      as a contextmanager::

          with Connection('host') as c:
             c.run('command')
             c.put('file')

      .. warning::
          While Fabric (and Paramiko) attempt to register connections for
          automatic garbage collection, it's not currently safe to rely on that
          feature, as it can lead to end-of-process hangs and similar behavior.

    .. note::
        This class rebinds `invoke.context.Context.run` to `.local` so both
        remote and local command execution can coexist.

    **Configuration**

    Most `.Connection` parameters honor :doc:`Invoke-style configuration
    </concepts/configuration>` as well as any applicable :ref:`SSH config file
    directives <connection-ssh-config>`. For example, to end up with a
    connection to ``admin@myhost``, one could:

    - Use any built-in config mechanism, such as ``/etc/fabric.yml``,
      ``~/.fabric.json``, collection-driven configuration, env vars, etc,
      stating ``user: admin`` (or ``{"user": "admin"}``, depending on config
      format.) Then ``Connection('myhost')`` would implicitly have a ``user``
      of ``admin``.
    - Use an SSH config file containing ``User admin`` within any applicable
      ``Host`` header (``Host myhost``, ``Host *``, etc.) Again,
      ``Connection('myhost')`` will default to an ``admin`` user.
    - Leverage host-parameter shorthand (described in `.Config.__init__`), i.e.
      ``Connection('admin@myhost')``.
    - Give the parameter directly: ``Connection('myhost', user='admin')``.

    The same applies to agent forwarding, gateways, and so forth.

    .. versionadded:: 2.0
    """

    # NOTE: these are initialized here to hint to invoke.Config.__setattr__
    # that they should be treated as real attributes instead of config proxies.
    # (Additionally, we're doing this instead of using invoke.Config._set() so
    # we can take advantage of Sphinx's attribute-doc-comment static analysis.)
    # Once an instance is created, these values will usually be non-None
    # because they default to the default config values.
    host = None
    original_host = None
    user = None
    port = None
    ssh_config = None
    gateway = None
    forward_agent = None
    connect_timeout = None
    connect_kwargs = None
    client = None
    transport = None
    _sftp = None
    _agent_handler = None

    @classmethod
    def from_v1(cls, env, **kwargs):
        """
        Alternate constructor which uses Fabric 1's ``env`` dict for settings.

        All keyword arguments besides ``env`` are passed unmolested into the
        primary constructor.

        .. warning::
            Because your own config overrides will win over data from ``env``,
            make sure you only set values you *intend* to change from your v1
            environment!

        For details on exactly which ``env`` vars are imported and what they
        become in the new API, please see :ref:`v1-env-var-imports`.

        :param env:
            An explicit Fabric 1 ``env`` dict (technically, any
            ``fabric.utils._AttributeDict`` instance should work) to pull
            configuration from.

        .. versionadded:: 2.4
        """
        # TODO: import fabric.state.env (need good way to test it first...)
        # TODO: how to handle somebody accidentally calling this in a process
        # where 'fabric' is fabric 2, and there's no fabric 1? Probably just a
        # re-raise of ImportError??
        # Our only requirement is a non-empty host_string
        if not env.host_string:
            raise InvalidV1Env(
                "Supplied v1 env has an empty `host_string` value! Please make sure you're calling Connection.from_v1 within a connected Fabric 1 session."  # noqa
            )
        # TODO: detect collisions with kwargs & except instead of overwriting?
        # (More Zen of Python compliant, but also, effort, and also, makes it
        # harder for users to intentionally overwrite!)
        connect_kwargs = kwargs.setdefault("connect_kwargs", {})
        kwargs.setdefault("host", env.host_string)
        shorthand = derive_shorthand(env.host_string)
        # TODO: don't we need to do the below skipping for user too?
        kwargs.setdefault("user", env.user)
        # Skip port if host string seemed to have it; otherwise we hit our own
        # ambiguity clause in __init__. v1 would also have been doing this
        # anyways (host string wins over other settings).
        if not shorthand["port"]:
            # Run port through int(); v1 inexplicably has a string default...
            kwargs.setdefault("port", int(env.port))
        # key_filename defaults to None in v1, but in v2, we expect it to be
        # either unset, or set to a list. Thus, we only pull it over if it is
        # not None.
        if env.key_filename is not None:
            connect_kwargs.setdefault("key_filename", env.key_filename)
        # Obtain config values, if not given, from its own from_v1
        # NOTE: not using setdefault as we truly only want to call
        # Config.from_v1 when necessary.
        if "config" not in kwargs:
            kwargs["config"] = Config.from_v1(env)
        return cls(**kwargs)

    # TODO: should "reopening" an existing Connection object that has been
    # closed, be allowed? (See e.g. how v1 detects closed/semi-closed
    # connections & nukes them before creating a new client to the same host.)
    # TODO: push some of this into paramiko.client.Client? e.g. expand what
    # Client.exec_command does, it already allows configuring a subset of what
    # we do / will eventually do / did in 1.x. It's silly to have to do
    # .get_transport().open_session().
    def __init__(
        self,
        host,
        user=None,
        port=None,
        config=None,
        gateway=None,
        forward_agent=None,
        connect_timeout=None,
        connect_kwargs=None,
        inline_ssh_env=None,
    ):
        """
        Set up a new object representing a server connection.

        :param str host:
            the hostname (or IP address) of this connection.

            May include shorthand for the ``user`` and/or ``port`` parameters,
            of the form ``user@host``, ``host:port``, or ``user@host:port``.

            .. note::
                Due to ambiguity, IPv6 host addresses are incompatible with the
                ``host:port`` shorthand (though ``user@host`` will still work
                OK). In other words, the presence of >1 ``:`` character will
                prevent any attempt to derive a shorthand port number; use the
                explicit ``port`` parameter instead.

            .. note::
                If ``host`` matches a ``Host`` clause in loaded SSH config
                data, and that ``Host`` clause contains a ``Hostname``
                directive, the resulting `.Connection` object will behave as if
                ``host`` is equal to that ``Hostname`` value.

                In all cases, the original value of ``host`` is preserved as
                the ``original_host`` attribute.

                Thus, given SSH config like so::

                    Host myalias
                        Hostname realhostname

                a call like ``Connection(host='myalias')`` will result in an
                object whose ``host`` attribute is ``realhostname``, and whose
                ``original_host`` attribute is ``myalias``.

        :param str user:
            the login user for the remote connection. Defaults to
            ``config.user``.

        :param int port:
            the remote port. Defaults to ``config.port``.

        :param config:
            configuration settings to use when executing methods on this
            `.Connection` (e.g. default SSH port and so forth).

            Should be a `.Config` or an `invoke.config.Config`
            (which will be turned into a `.Config`).

            Default is an anonymous `.Config` object.

        :param gateway:
            An object to use as a proxy or gateway for this connection.

            This parameter accepts one of the following:

            - another `.Connection` (for a ``ProxyJump`` style gateway);
            - a shell command string (for a ``ProxyCommand`` style style
              gateway).

            Default: ``None``, meaning no gatewaying will occur (unless
            otherwise configured; if one wants to override a configured gateway
            at runtime, specify ``gateway=False``.)

            .. seealso:: :ref:`ssh-gateways`

        :param bool forward_agent:
            Whether to enable SSH agent forwarding.

            Default: ``config.forward_agent``.

        :param int connect_timeout:
            Connection timeout, in seconds.

            Default: ``config.timeouts.connect``.


        :param dict connect_kwargs:

            .. _connect_kwargs-arg:

            Keyword arguments handed verbatim to
            `SSHClient.connect <paramiko.client.SSHClient.connect>` (when
            `.open` is called).

            `.Connection` tries not to grow additional settings/kwargs of its
            own unless it is adding value of some kind; thus,
            ``connect_kwargs`` is currently the right place to hand in paramiko
            connection parameters such as ``pkey`` or ``key_filename``. For
            example::

                c = Connection(
                    host="hostname",
                    user="admin",
                    connect_kwargs={
                        "key_filename": "/home/myuser/.ssh/private.key",
                    },
                )

            Default: ``config.connect_kwargs``.

        :param bool inline_ssh_env:
            Whether to send environment variables "inline" as prefixes in front
            of command strings (``export VARNAME=value && mycommand here``;
            this is the default behavior), or submit them through the SSH
            protocol itself.

            In Fabric 2.x this defaulted to ``False`` (try using the protocol
            behavior), but in 3.x it changed to ``True`` due to the simple fact
            that most remote servers are deployed with a restricted
            ``AcceptEnv`` setting, making use of the protocol approach
            non-viable.

            The actual default value is the value of the ``inline_ssh_env``
            :ref:`configuration value <default-values>` (which, as above,
            currently defaults to ``True``).

            .. warning::
                This functionality does **not** currently perform any shell
                escaping on your behalf! Be careful when using nontrivial
                values, and note that you can put in your own quoting,
                backslashing etc if desired.

                Consider using a different approach (such as actual
                remote shell scripts) if you run into too many issues here.

            .. note::
                When serializing into prefixed ``FOO=bar`` format, we apply the
                builtin `sorted` function to the env dictionary's keys, to
                remove what would otherwise be ambiguous/arbitrary ordering.

            .. note::
                This setting has no bearing on *local* shell commands; it only
                affects remote commands, and thus, methods like `.run` and
                `.sudo`.

        :raises ValueError:
            if user or port values are given via both ``host`` shorthand *and*
            their own arguments. (We `refuse the temptation to guess`_).

        .. _refuse the temptation to guess:
            http://zen-of-python.info/
            in-the-face-of-ambiguity-refuse-the-temptation-to-guess.html#12

        .. versionchanged:: 2.3
            Added the ``inline_ssh_env`` parameter.

        .. versionchanged:: 3.0
            ``inline_ssh_env`` still defaults to the config value, but said
            config value has now changed and defaults to ``True``, not
            ``False``.
        """
        # NOTE: parent __init__ sets self._config; for now we simply overwrite
        # that below. If it's somehow problematic we would want to break parent
        # __init__ up in a manner that is more cleanly overrideable.
        super().__init__(config=config)

        #: The .Config object referenced when handling default values (for e.g.
        #: user or port, when not explicitly given) or deciding how to behave.
        if config is None:
            config = Config()
        # Handle 'vanilla' Invoke config objects, which need cloning 'into' one
        # of our own Configs (which grants the new defaults, etc, while not
        # squashing them if the Invoke-level config already accounted for them)
        elif not isinstance(config, Config):
            config = config.clone(into=Config)
        self._set(_config=config)
        # TODO: when/how to run load_files, merge, load_shell_env, etc?
        # TODO: i.e. what is the lib use case here (and honestly in invoke too)

        shorthand = self.derive_shorthand(host)
        host = shorthand["host"]
        err = "You supplied the {} via both shorthand and kwarg! Please pick one."  # noqa
        if shorthand["user"] is not None:
            if user is not None:
                raise ValueError(err.format("user"))
            user = shorthand["user"]
        if shorthand["port"] is not None:
            if port is not None:
                raise ValueError(err.format("port"))
            port = shorthand["port"]

        # NOTE: we load SSH config data as early as possible as it has
        # potential to affect nearly every other attribute.
        #: The per-host SSH config data, if any. (See :ref:`ssh-config`.)
        self.ssh_config = self.config.base_ssh_config.lookup(host)

        self.original_host = host
        #: The hostname of the target server.
        self.host = host
        if "hostname" in self.ssh_config:
            # TODO: log that this occurred?
            self.host = self.ssh_config["hostname"]

        #: The username this connection will use to connect to the remote end.
        self.user = user or self.ssh_config.get("user", self.config.user)
        # TODO: is it _ever_ possible to give an empty user value (e.g.
        # user='')? E.g. do some SSH server specs allow for that?

        #: The network port to connect on.
        self.port = port or int(self.ssh_config.get("port", self.config.port))

        # Gateway/proxy/bastion/jump setting: non-None values - string,
        # Connection, even eg False - get set directly; None triggers seek in
        # config/ssh_config
        #: The gateway `.Connection` or ``ProxyCommand`` string to be used,
        #: if any.
        self.gateway = gateway if gateway is not None else self.get_gateway()
        # NOTE: we use string above, vs ProxyCommand obj, to avoid spinning up
        # the ProxyCommand subprocess at init time, vs open() time.
        # TODO: make paramiko.proxy.ProxyCommand lazy instead?

        if forward_agent is None:
            # Default to config...
            forward_agent = self.config.forward_agent
            # But if ssh_config is present, it wins
            if "forwardagent" in self.ssh_config:
                # TODO: SSHConfig really, seriously needs some love here, god
                map_ = {"yes": True, "no": False}
                forward_agent = map_[self.ssh_config["forwardagent"]]
        #: Whether agent forwarding is enabled.
        self.forward_agent = forward_agent

        if connect_timeout is None:
            connect_timeout = self.ssh_config.get(
                "connecttimeout", self.config.timeouts.connect
            )
        if connect_timeout is not None:
            connect_timeout = int(connect_timeout)
        #: Connection timeout
        self.connect_timeout = connect_timeout

        #: Keyword arguments given to `paramiko.client.SSHClient.connect` when
        #: `open` is called.
        self.connect_kwargs = self.resolve_connect_kwargs(connect_kwargs)

        #: The `paramiko.client.SSHClient` instance this connection wraps.
        client = SSHClient()
        client.set_missing_host_key_policy(AutoAddPolicy())
        self.client = client

        #: A convenience handle onto the return value of
        #: ``self.client.get_transport()`` (after connection time).
        self.transport = None

        if inline_ssh_env is None:
            inline_ssh_env = self.config.inline_ssh_env
        #: Whether to construct remote command lines with env vars prefixed
        #: inline.
        self.inline_ssh_env = inline_ssh_env

    def resolve_connect_kwargs(self, connect_kwargs):
        # TODO: is it better to pre-empt conflicts w/ manually-handled
        # connect() kwargs (hostname, username, etc) here or in open()? We're
        # doing open() for now in case e.g. someone manually modifies
        # .connect_kwargs attributewise, but otherwise it feels better to do it
        # early instead of late.
        constructor_kwargs = connect_kwargs or {}
        config_kwargs = self.config.connect_kwargs
        constructor_keys = constructor_kwargs.get("key_filename", [])
        config_keys = config_kwargs.get("key_filename", [])
        ssh_config_keys = self.ssh_config.get("identityfile", [])

        # Default data: constructor if given, config otherwise
        final_kwargs = constructor_kwargs or config_kwargs

        # Key filename: merge, in order, config (which includes CLI flags),
        # then constructor kwargs, and finally SSH config file data.
        # Make sure all are normalized to list as well!
        final_keys = []
        for value in (config_keys, constructor_keys, ssh_config_keys):
            if isinstance(value, str):
                value = [value]
            final_keys.extend(value)
        # Only populate if non-empty.
        if final_keys:
            final_kwargs["key_filename"] = final_keys

        return final_kwargs

    def get_gateway(self):
        # SSH config wins over Invoke-style config
        if "proxyjump" in self.ssh_config:
            # Reverse hop1,hop2,hop3 style ProxyJump directive so we start
            # with the final (itself non-gatewayed) hop and work up to
            # the front (actual, supplied as our own gateway) hop
            hops = reversed(self.ssh_config["proxyjump"].split(","))
            prev_gw = None
            for hop in hops:
                # Short-circuit if we appear to be our own proxy, which would
                # be a RecursionError. Implies SSH config wildcards.
                # TODO: in an ideal world we'd check user/port too in case they
                # differ, but...seriously? They can file a PR with those extra
                # half dozen test cases in play, E_NOTIME
                if self.derive_shorthand(hop)["host"] == self.host:
                    return None
                # Happily, ProxyJump uses identical format to our host
                # shorthand...
                kwargs = dict(config=self.config.clone())
                if prev_gw is not None:
                    kwargs["gateway"] = prev_gw
                cxn = Connection(hop, **kwargs)
                prev_gw = cxn
            return prev_gw
        elif "proxycommand" in self.ssh_config:
            # Just a string, which we interpret as a proxy command..
            return self.ssh_config["proxycommand"]
        # Fallback: config value (may be None).
        return self.config.gateway

    def __repr__(self):
        # Host comes first as it's the most common differentiator by far
        bits = [("host", self.host)]
        # TODO: maybe always show user regardless? Explicit is good...
        if self.user != self.config.user:
            bits.append(("user", self.user))
        # TODO: harder to make case for 'always show port'; maybe if it's
        # non-22 (even if config has overridden the local default)?
        if self.port != self.config.port:
            bits.append(("port", self.port))
        # NOTE: sometimes self.gateway may be eg False if someone wants to
        # explicitly override a configured non-None value (as otherwise it's
        # impossible for __init__ to tell if a None means "nothing given" or
        # "seriously please no gatewaying". So, this must always be a vanilla
        # truth test and not eg "is not None".
        if self.gateway:
            # Displaying type because gw params would probs be too verbose
            val = "proxyjump"
            if isinstance(self.gateway, str):
                val = "proxycommand"
            bits.append(("gw", val))
        return "<Connection {}>".format(
            " ".join("{}={}".format(*x) for x in bits)
        )

    def _identity(self):
        # TODO: consider including gateway and maybe even other init kwargs?
        # Whether two cxns w/ same user/host/port but different
        # gateway/keys/etc, should be considered "the same", is unclear.
        return (self.host, self.user, self.port)

    def __eq__(self, other):
        if not isinstance(other, Connection):
            return False
        return self._identity() == other._identity()

    def __lt__(self, other):
        return self._identity() < other._identity()

    def __hash__(self):
        # NOTE: this departs from Context/DataProxy, which is not usefully
        # hashable.
        return hash(self._identity())

    def derive_shorthand(self, host_string):
        # NOTE: used to be defined inline; preserving API call for both
        # backwards compatibility and because it seems plausible we may want to
        # modify behavior later, using eg config or other attributes.
        return derive_shorthand(host_string)

    @property
    def is_connected(self):
        """
        Whether or not this connection is actually open.

        .. versionadded:: 2.0
        """
        return self.transport.active if self.transport else False

    def open(self):
        """
        Initiate an SSH connection to the host/port this object is bound to.

        This may include activating the configured gateway connection, if one
        is set.

        Also saves a handle to the now-set Transport object for easier access.

        Various connect-time settings (and/or their corresponding :ref:`SSH
        config options <ssh-config>`) are utilized here in the call to
        `SSHClient.connect <paramiko.client.SSHClient.connect>`. (For details,
        see :doc:`the configuration docs </concepts/configuration>`.)

        :returns:
            The result of the internal call to `.SSHClient.connect`, if
            performing an initial connection; ``None`` otherwise.

        .. versionadded:: 2.0
        .. versionchanged:: 3.1
            Now returns the inner Paramiko connect call's return value instead
            of always returning the implicit ``None``.
        """
        # Short-circuit
        if self.is_connected:
            return
        err = "Refusing to be ambiguous: connect() kwarg '{}' was given both via regular arg and via connect_kwargs!"  # noqa
        # These may not be given, period
        for key in """
            hostname
            port
            username
        """.split():
            if key in self.connect_kwargs:
                raise ValueError(err.format(key))
        # These may be given one way or the other, but not both
        if (
            "timeout" in self.connect_kwargs
            and self.connect_timeout is not None
        ):
            raise ValueError(err.format("timeout"))
        # No conflicts -> merge 'em together
        kwargs = dict(
            self.connect_kwargs,
            username=self.user,
            hostname=self.host,
            port=self.port,
        )
        if self.gateway:
            kwargs["sock"] = self.open_gateway()
        if self.connect_timeout:
            kwargs["timeout"] = self.connect_timeout
        # Strip out empty defaults for less noisy debugging
        if "key_filename" in kwargs and not kwargs["key_filename"]:
            del kwargs["key_filename"]
        auth_strategy_class = self.authentication.strategy_class
        if auth_strategy_class is not None:
            # Pop connect_kwargs related to auth to avoid giving Paramiko
            # conflicting signals.
            for key in (
                "allow_agent",
                "key_filename",
                "look_for_keys",
                "passphrase",
                "password",
                "pkey",
                "username",
            ):
                kwargs.pop(key, None)

            kwargs["auth_strategy"] = auth_strategy_class(
                ssh_config=self.ssh_config,
                fabric_config=self.config,
                username=self.user,
            )
        # Actually connect!
        result = self.client.connect(**kwargs)
        self.transport = self.client.get_transport()
        return result

    def open_gateway(self):
        """
        Obtain a socket-like object from `gateway`.

        :returns:
            A ``direct-tcpip`` `paramiko.channel.Channel`, if `gateway` was a
            `.Connection`; or a `~paramiko.proxy.ProxyCommand`, if `gateway`
            was a string.

        .. versionadded:: 2.0
        """
        # ProxyCommand is faster to set up, so do it first.
        if isinstance(self.gateway, str):
            # Leverage a dummy SSHConfig to ensure %h/%p/etc are parsed.
            # TODO: use real SSH config once loading one properly is
            # implemented.
            ssh_conf = SSHConfig()
            dummy = "Host {}\n    ProxyCommand {}"
            ssh_conf.parse(StringIO(dummy.format(self.host, self.gateway)))
            return ProxyCommand(ssh_conf.lookup(self.host)["proxycommand"])
        # Handle inner-Connection gateway type here.
        # TODO: logging
        self.gateway.open()
        # TODO: expose the opened channel itself as an attribute? (another
        # possible argument for separating the two gateway types...) e.g. if
        # someone wanted to piggyback on it for other same-interpreter socket
        # needs...
        # TODO: and the inverse? allow users to supply their own socket/like
        # object they got via $WHEREEVER?
        # TODO: how best to expose timeout param? reuse general connection
        # timeout from config?
        return self.gateway.transport.open_channel(
            kind="direct-tcpip",
            dest_addr=(self.host, int(self.port)),
            # NOTE: src_addr needs to be 'empty but not None' values to
            # correctly encode into a network message. Theoretically Paramiko
            # could auto-interpret None sometime & save us the trouble.
            src_addr=("", 0),
        )

    def close(self):
        """
        Terminate the network connection to the remote end, if open.

        If any SFTP sessions are open, they will also be closed.

        If no connection or SFTP session is open, this method does nothing.

        .. versionadded:: 2.0
        .. versionchanged:: 3.0
            Now closes SFTP sessions too (2.x required manually doing so).
        """
        if self._sftp is not None:
            self._sftp.close()
            self._sftp = None

        if self.is_connected:
            self.client.close()
            if self.forward_agent and self._agent_handler is not None:
                self._agent_handler.close()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    @opens
    def create_session(self):
        channel = self.transport.open_session()
        if self.forward_agent:
            self._agent_handler = AgentRequestHandler(channel)
        return channel

    def _remote_runner(self):
        return self.config.runners.remote(
            context=self, inline_env=self.inline_ssh_env
        )

    @opens
    def run(self, command, **kwargs):
        """
        Execute a shell command on the remote end of this connection.

        This method wraps an SSH-capable implementation of
        `invoke.runners.Runner.run`; see its documentation for details.

        .. warning::
            There are a few spots where Fabric departs from Invoke's default
            settings/behaviors; they are documented under
            `.Config.global_defaults`.

        .. versionadded:: 2.0
        """
        return self._run(self._remote_runner(), command, **kwargs)

    @opens
    def sudo(self, command, **kwargs):
        """
        Execute a shell command, via ``sudo``, on the remote end.

        This method is identical to `invoke.context.Context.sudo` in every way,
        except in that -- like `run` -- it honors per-host/per-connection
        configuration overrides in addition to the generic/global ones. Thus,
        for example, per-host sudo passwords may be configured.

        .. versionadded:: 2.0
        """
        return self._sudo(self._remote_runner(), command, **kwargs)

    @opens
    def shell(self, **kwargs):
        """
        Run an interactive login shell on the remote end, as with ``ssh``.

        This method is intended strictly for use cases where you can't know
        what remote shell to invoke, or are connecting to a non-POSIX-server
        environment such as a network appliance or other custom SSH server.
        Nearly every other use case, including interactively-focused ones, will
        be better served by using `run` plus an explicit remote shell command
        (eg ``bash``).

        `shell` has the following differences in behavior from `run`:

        - It still returns a `~invoke.runners.Result` instance, but the object
          will have a less useful set of attributes than with `run` or `local`:

            - ``command`` will be ``None``, as there is no such input argument.
            - ``stdout`` will contain a full record of the session, including
              all interactive input, as that is echoed back to the user. This
              can be useful for logging but is much less so for doing
              programmatic things after the method returns.
            - ``stderr`` will always be empty (same as `run` when
              ``pty==True``).
            - ``pty`` will always be True (because one was automatically used).
            - ``exited`` and similar attributes will only reflect the overall
              session, which may vary by shell or appliance but often has no
              useful relationship with the internally executed commands' exit
              codes.

        - This method behaves as if ``warn`` is set to ``True``: even if the
          remote shell exits uncleanly, no exception will be raised.
        - A pty is always allocated remotely, as with ``pty=True`` under `run`.
        - The ``inline_env`` setting is ignored, as there is no default shell
          command to add the parameters to (and no guarantee the remote end
          even is a shell!)

        It supports **only** the following kwargs, which behave identically to
        their counterparts in `run` unless otherwise stated:

        - ``encoding``
        - ``env``
        - ``in_stream`` (useful in niche cases, but make sure regular `run`
          with this argument isn't more suitable!)
        - ``replace_env``
        - ``watchers`` (note that due to pty echoing your stdin back to stdout,
          a watcher will see your input as well as program stdout!)

        Those keyword arguments also honor the ``run.*`` configuration tree, as
        in `run`/`sudo`.

        :returns: `~invoke.runners.Result`

        :raises:
            `~invoke.exceptions.ThreadException` (if the background I/O threads
            encountered exceptions other than
            `~invoke.exceptions.WatcherError`).

        .. versionadded:: 2.7
        """
        runner = self.config.runners.remote_shell(context=self)
        # Reinstate most defaults as explicit kwargs to ensure user's config
        # doesn't make this mode break horribly. Then override a few that need
        # to change, like pty.
        allowed = ("encoding", "env", "in_stream", "replace_env", "watchers")
        new_kwargs = {}
        for key, value in self.config.global_defaults()["run"].items():
            if key in allowed:
                # Use allowed kwargs if given, otherwise also fill them from
                # defaults
                new_kwargs[key] = kwargs.pop(key, self.config.run[key])
            else:
                new_kwargs[key] = value
        new_kwargs.update(pty=True)
        # At this point, any leftover kwargs would be ignored, so yell instead
        if kwargs:
            err = "shell() got unexpected keyword arguments: {!r}"
            raise TypeError(err.format(list(kwargs.keys())))
        return runner.run(command=None, **new_kwargs)

    def local(self, *args, **kwargs):
        """
        Execute a shell command on the local system.

        This method is effectively a wrapper of `invoke.run`; see its docs for
        details and call signature.

        .. versionadded:: 2.0
        """
        # Superclass run() uses runners.local, so we can literally just call it
        # straight.
        return super().run(*args, **kwargs)

    @opens
    def sftp(self):
        """
        Return a `~paramiko.sftp_client.SFTPClient` object.

        If called more than one time, memoizes the first result; thus, any
        given `.Connection` instance will only ever have a single SFTP client,
        and state (such as that managed by
        `~paramiko.sftp_client.SFTPClient.chdir`) will be preserved.

        .. versionadded:: 2.0
        """
        if self._sftp is None:
            self._sftp = self.client.open_sftp()
        return self._sftp

    def get(self, *args, **kwargs):
        """
        Get a remote file to the local filesystem or file-like object.

        Simply a wrapper for `.Transfer.get`. Please see its documentation for
        all details.

        .. versionadded:: 2.0
        """
        return Transfer(self).get(*args, **kwargs)

    def put(self, *args, **kwargs):
        """
        Put a local file (or file-like object) to the remote filesystem.

        Simply a wrapper for `.Transfer.put`. Please see its documentation for
        all details.

        .. versionadded:: 2.0
        """
        return Transfer(self).put(*args, **kwargs)

    # TODO: yield the socket for advanced users? Other advanced use cases
    # (perhaps factor out socket creation itself)?
    # TODO: probably push some of this down into Paramiko
    @contextmanager
    @opens
    def forward_local(
        self,
        local_port,
        remote_port=None,
        remote_host="localhost",
        local_host="localhost",
    ):
        """
        Open a tunnel connecting ``local_port`` to the server's environment.

        For example, say you want to connect to a remote PostgreSQL database
        which is locked down and only accessible via the system it's running
        on. You have SSH access to this server, so you can temporarily make
        port 5432 on your local system act like port 5432 on the server::

            import psycopg2
            from fabric import Connection

            with Connection('my-db-server').forward_local(5432):
                db = psycopg2.connect(
                    host='localhost', port=5432, database='mydb'
                )
                # Do things with 'db' here

        This method is analogous to using the ``-L`` option of OpenSSH's
        ``ssh`` program.

        :param int local_port: The local port number on which to listen.

        :param int remote_port:
            The remote port number. Defaults to the same value as
            ``local_port``.

        :param str local_host:
            The local hostname/interface on which to listen. Default:
            ``localhost``.

        :param str remote_host:
            The remote hostname serving the forwarded remote port. Default:
            ``localhost`` (i.e., the host this `.Connection` is connected to.)

        :returns:
            Nothing; this method is only useful as a context manager affecting
            local operating system state.

        .. versionadded:: 2.0
        """
        if not remote_port:
            remote_port = local_port

        # TunnelManager does all of the work, sitting in the background (so we
        # can yield) and spawning threads every time somebody connects to our
        # local port.
        finished = Event()
        manager = TunnelManager(
            local_port=local_port,
            local_host=local_host,
            remote_port=remote_port,
            remote_host=remote_host,
            # TODO: not a huge fan of handing in our transport, but...?
            transport=self.transport,
            finished=finished,
        )
        manager.start()

        # Return control to caller now that things ought to be operational
        try:
            yield
        # Teardown once user exits block
        finally:
            # Signal to manager that it should close all open tunnels
            finished.set()
            # Then wait for it to do so
            manager.join()
            # Raise threading errors from within the manager, which would be
            # one of:
            # - an inner ThreadException, which was created by the manager on
            # behalf of its Tunnels; this gets directly raised.
            # - some other exception, which would thus have occurred in the
            # manager itself; we wrap this in a new ThreadException.
            # NOTE: in these cases, some of the metadata tracking in
            # ExceptionHandlingThread/ExceptionWrapper/ThreadException (which
            # is useful when dealing with multiple nearly-identical sibling IO
            # threads) is superfluous, but it doesn't feel worth breaking
            # things up further; we just ignore it for now.
            wrapper = manager.exception()
            if wrapper is not None:
                if wrapper.type is ThreadException:
                    raise wrapper.value
                else:
                    raise ThreadException([wrapper])

            # TODO: cancel port forward on transport? Does that even make sense
            # here (where we used direct-tcpip) vs the opposite method (which
            # is what uses forward-tcpip)?

    # TODO: probably push some of this down into Paramiko
    @contextmanager
    @opens
    def forward_remote(
        self,
        remote_port,
        local_port=None,
        remote_host="127.0.0.1",
        local_host="localhost",
    ):
        """
        Open a tunnel connecting ``remote_port`` to the local environment.

        For example, say you're running a daemon in development mode on your
        workstation at port 8080, and want to funnel traffic to it from a
        production or staging environment.

        In most situations this isn't possible as your office/home network
        probably blocks inbound traffic. But you have SSH access to this
        server, so you can temporarily make port 8080 on that server act like
        port 8080 on your workstation::

            from fabric import Connection

            c = Connection('my-remote-server')
            with c.forward_remote(8080):
                c.run("remote-data-writer --port 8080")
                # Assuming remote-data-writer runs until interrupted, this will
                # stay open until you Ctrl-C...

        This method is analogous to using the ``-R`` option of OpenSSH's
        ``ssh`` program.

        :param int remote_port: The remote port number on which to listen.

        :param int local_port:
            The local port number. Defaults to the same value as
            ``remote_port``.

        :param str local_host:
            The local hostname/interface the forwarded connection talks to.
            Default: ``localhost``.

        :param str remote_host:
            The remote interface address to listen on when forwarding
            connections. Default: ``127.0.0.1`` (i.e. only listen on the remote
            localhost).

        :returns:
            Nothing; this method is only useful as a context manager affecting
            local operating system state.

        .. versionadded:: 2.0
        """
        if not local_port:
            local_port = remote_port
        # Callback executes on each connection to the remote port and is given
        # a Channel hooked up to said port. (We don't actually care about the
        # source/dest host/port pairs at all; only whether the channel has data
        # to read and suchlike.)
        # We then pair that channel with a new 'outbound' socket connection to
        # the local host/port being forwarded, in a new Tunnel.
        # That Tunnel is then added to a shared data structure so we can track
        # & close them during shutdown.
        #
        # TODO: this approach is less than ideal because we have to share state
        # between ourselves & the callback handed into the transport's own
        # thread handling (which is roughly analogous to our self-controlled
        # TunnelManager for local forwarding). See if we can use more of
        # Paramiko's API (or improve it and then do so) so that isn't
        # necessary.
        tunnels = []

        def callback(channel, src_addr_tup, dst_addr_tup):
            sock = socket.socket()
            # TODO: handle connection failure such that channel, etc get closed
            sock.connect((local_host, local_port))
            # TODO: we don't actually need to generate the Events at our level,
            # do we? Just let Tunnel.__init__ do it; all we do is "press its
            # button" on shutdown...
            tunnel = Tunnel(channel=channel, sock=sock, finished=Event())
            tunnel.start()
            # Communication between ourselves & the Paramiko handling subthread
            tunnels.append(tunnel)

        # Ask Paramiko (really, the remote sshd) to call our callback whenever
        # connections are established on the remote iface/port.
        # transport.request_port_forward(remote_host, remote_port, callback)
        try:
            self.transport.request_port_forward(
                address=remote_host, port=remote_port, handler=callback
            )
            yield
        finally:
            # TODO: see above re: lack of a TunnelManager
            # TODO: and/or also refactor with TunnelManager re: shutdown logic.
            # E.g. maybe have a non-thread TunnelManager-alike with a method
            # that acts as the callback? At least then there's a tiny bit more
            # encapsulation...meh.
            for tunnel in tunnels:
                tunnel.finished.set()
                tunnel.join()
            self.transport.cancel_port_forward(
                address=remote_host, port=remote_port
            )



from unittest.mock import Mock, patch, call
from pytest import mark, raises

from fabric import Connection, Group, SerialGroup, ThreadingGroup, GroupResult
from fabric.group import thread_worker
from fabric.exceptions import GroupException


RUNNER_METHODS = ("run", "sudo")
TRANSFER_METHODS = ("put", "get")
ALL_METHODS = RUNNER_METHODS + TRANSFER_METHODS
runner_args = ("command",)
runner_kwargs = dict(hide=True, warn=True)
transfer_args = tuple()
transfer_kwargs = dict(local="yokel", remote="goat")
ARGS_BY_METHOD = dict(
    run=runner_args, sudo=runner_args, put=transfer_args, get=transfer_args
)
KWARGS_BY_METHOD = dict(
    run=runner_kwargs,
    sudo=runner_kwargs,
    put=transfer_kwargs,
    get=transfer_kwargs,
)


class Group_:
    class init:
        "__init__"

        def may_be_empty(self):
            assert len(Group()) == 0

        def takes_splat_arg_of_host_strings(self):
            g = Group("foo", "bar")
            assert g[0].host == "foo"
            assert g[1].host == "bar"

        def takes_splat_kwargs_and_passes_them_to_Connections(self):
            g = Group("foo", "bar", user="admin", forward_agent=True)
            assert g[0].host == "foo"
            assert g[0].user == "admin"
            assert g[0].forward_agent is True
            assert g[1].host == "bar"
            assert g[1].user == "admin"
            assert g[1].forward_agent is True

    class from_connections:
        def inits_from_iterable_of_Connections(self):
            g = Group.from_connections((Connection("foo"), Connection("bar")))
            assert len(g) == 2
            assert g[1].host == "bar"

    def acts_like_an_iterable_of_Connections(self):
        g = Group("foo", "bar", "biz")
        assert g[0].host == "foo"
        assert g[-1].host == "biz"
        assert len(g) == 3
        for c in g:
            assert isinstance(c, Connection)

    @mark.parametrize("method", ALL_METHODS)
    def abstract_methods_not_implemented(self, method):
        group = Group()
        with raises(NotImplementedError):
            getattr(group, method)()

    class close_and_contextmanager_behavior:
        def close_closes_all_member_connections(self):
            cxns = [Mock(name=x) for x in ("foo", "bar", "biz")]
            g = Group.from_connections(cxns)
            g.close()
            for c in cxns:
                c.close.assert_called_once_with()

        def contextmanager_behavior_works_like_Connection(self):
            cxns = [Mock(name=x) for x in ("foo", "bar", "biz")]
            g = Group.from_connections(cxns)
            with g as my_g:
                assert my_g is g
            for c in cxns:
                c.close.assert_called_once_with()

    class get:
        class local_defaults_to_host_interpolated_path:
            def when_no_arg_or_kwarg_given(self):
                g = Group("host1", "host2")
                g._do = Mock()
                g.get(remote="whatever")
                g._do.assert_called_with(
                    "get", remote="whatever", local="{host}/"
                )

            def not_when_arg_given(self):
                g = Group("host1", "host2")
                g._do = Mock()
                g.get("whatever", "lol")
                # No local kwarg passed.
                g._do.assert_called_with("get", "whatever", "lol")

            def not_when_kwarg_given(self):
                g = Group("host1", "host2")
                g._do = Mock()
                g.get(remote="whatever", local="lol")
                # Doesn't stomp given local arg
                g._do.assert_called_with("get", remote="whatever", local="lol")


def _make_serial_tester(method, cxns, index, args, kwargs):
    args = args[:]
    kwargs = kwargs.copy()

    def tester(*a, **k):  # Don't care about doing anything with our own args.
        car, cdr = index, index + 1
        predecessors = cxns[:car]
        successors = cxns[cdr:]
        for predecessor in predecessors:
            getattr(predecessor, method).assert_called_with(*args, **kwargs)
        for successor in successors:
            assert not getattr(successor, method).called

    return tester


class SerialGroup_:
    @mark.parametrize("method", ALL_METHODS)
    def executes_arguments_on_contents_run_serially(self, method):
        "executes arguments on contents' run() serially"
        cxns = [Connection(x) for x in ("host1", "host2", "host3")]
        args = ARGS_BY_METHOD[method]
        kwargs = KWARGS_BY_METHOD[method]
        for index, cxn in enumerate(cxns):
            side_effect = _make_serial_tester(
                method, cxns, index, args, kwargs
            )
            setattr(cxn, method, Mock(side_effect=side_effect))
        g = SerialGroup.from_connections(cxns)
        getattr(g, method)(*args, **kwargs)
        # Sanity check, e.g. in case none of them were actually run
        for cxn in cxns:
            getattr(cxn, method).assert_called_with(*args, **kwargs)

    @mark.parametrize("method", ALL_METHODS)
    def errors_in_execution_capture_and_continue_til_end(self, method):
        cxns = [Mock(name=x) for x in ("host1", "host2", "host3")]

        class OhNoz(Exception):
            pass

        onoz = OhNoz()
        getattr(cxns[1], method).side_effect = onoz
        g = SerialGroup.from_connections(cxns)
        try:
            getattr(g, method)("whatever", hide=True)
        except GroupException as e:
            result = e.result
        else:
            assert False, "Did not raise GroupException!"
        succeeded = {
            cxns[0]: getattr(cxns[0], method).return_value,
            cxns[2]: getattr(cxns[2], method).return_value,
        }
        failed = {cxns[1]: onoz}
        expected = succeeded.copy()
        expected.update(failed)
        assert result == expected
        assert result.succeeded == succeeded
        assert result.failed == failed

    @mark.parametrize("method", ALL_METHODS)
    def returns_results_mapping(self, method):
        cxns = [Mock(name=x) for x in ("host1", "host2", "host3")]
        g = SerialGroup.from_connections(cxns)
        result = getattr(g, method)("whatever", hide=True)
        assert isinstance(result, GroupResult)
        expected = {x: getattr(x, method).return_value for x in cxns}
        assert result == expected
        assert result.succeeded == expected
        assert result.failed == {}


class ThreadingGroup_:
    def setup(self):
        self.cxns = [Connection(x) for x in ("host1", "host2", "host3")]

    @mark.parametrize("method", ALL_METHODS)
    @patch("fabric.group.Queue")
    @patch("fabric.group.ExceptionHandlingThread")
    def executes_arguments_on_contents_run_via_threading(
        self, Thread, Queue, method
    ):
        queue = Queue.return_value
        g = ThreadingGroup.from_connections(self.cxns)
        # Make sure .exception() doesn't yield truthy Mocks. Otherwise we
        # end up with 'exceptions' that cause errors due to all being the
        # same.
        Thread.return_value.exception.return_value = None
        args = ARGS_BY_METHOD[method]
        kwargs = KWARGS_BY_METHOD[method]
        getattr(g, method)(*args, **kwargs)
        # Testing that threads were used the way we expect is mediocre but
        # I honestly can't think of another good way to assert "threading
        # was used & concurrency occurred"...
        instantiations = [
            call(
                target=thread_worker,
                kwargs=dict(
                    cxn=cxn,
                    queue=queue,
                    method=method,
                    args=args,
                    kwargs=kwargs,
                ),
            )
            for cxn in self.cxns
        ]
        Thread.assert_has_calls(instantiations, any_order=True)
        # These ought to work as by default a Mock.return_value is a
        # singleton mock object
        expected = len(self.cxns)
        for name, got in (
            ("start", Thread.return_value.start.call_count),
            ("join", Thread.return_value.join.call_count),
        ):
            err = "Expected {} calls to ExceptionHandlingThread.{}, got {}"  # noqa
            err = err.format(expected, name, got)
            assert expected, got == err

    @mark.parametrize("method", ALL_METHODS)
    @patch("fabric.group.Queue")
    def queue_used_to_return_results(self, Queue, method):
        # Regular, explicit, mocks for Connections
        cxns = [Mock(host=x) for x in ("host1", "host2", "host3")]
        # Set up Queue with enough behavior to work / assert
        queue = Queue.return_value
        # Ending w/ a True will terminate a while-not-empty loop
        queue.empty.side_effect = (False, False, False, True)
        fakes = [(x, getattr(x, method).return_value) for x in cxns]
        queue.get.side_effect = fakes[:]
        # Execute & inspect results
        g = ThreadingGroup.from_connections(cxns)
        results = getattr(g, method)(
            *ARGS_BY_METHOD[method], **KWARGS_BY_METHOD[method]
        )
        expected = {x: getattr(x, method).return_value for x in cxns}
        assert results == expected
        # Make sure queue was used as expected within worker &
        # ThreadingGroup.run()
        puts = [call(x) for x in fakes]
        queue.put.assert_has_calls(puts, any_order=True)
        assert queue.empty.called
        gets = [call(block=False) for _ in cxns]
        queue.get.assert_has_calls(gets)

    @mark.parametrize("method", ALL_METHODS)
    def bubbles_up_errors_within_threads(self, method):
        # TODO: I feel like this is the first spot where a raw
        # ThreadException might need tweaks, at least presentation-wise,
        # since we're no longer dealing with truly background threads (IO
        # workers and tunnels), but "middle-ground" threads the user is
        # kind of expecting (and which they might expect to encounter
        # failures).
        cxns = [Mock(host=x) for x in ("host1", "host2", "host3")]

        class OhNoz(Exception):
            pass

        onoz = OhNoz()
        getattr(cxns[1], method).side_effect = onoz
        g = ThreadingGroup.from_connections(cxns)
        try:
            getattr(g, method)(
                *ARGS_BY_METHOD[method], **KWARGS_BY_METHOD[method]
            )
        except GroupException as e:
            result = e.result
        else:
            assert False, "Did not raise GroupException!"
        succeeded = {
            cxns[0]: getattr(cxns[0], method).return_value,
            cxns[2]: getattr(cxns[2], method).return_value,
        }
        failed = {cxns[1]: onoz}
        expected = succeeded.copy()
        expected.update(failed)
        assert result == expected
        assert result.succeeded == succeeded
        assert result.failed == failed

    @mark.parametrize("method", ALL_METHODS)
    def returns_results_mapping(self, method):
        cxns = [Mock(name=x) for x in ("host1", "host2", "host3")]
        g = ThreadingGroup.from_connections(cxns)
        result = getattr(g, method)("whatever", hide=True)
        assert isinstance(result, GroupResult)
        expected = {x: getattr(x, method).return_value for x in cxns}
        assert result == expected
        assert result.succeeded == expected
        assert result.failed == {}



"""
Tests testing the fabric.util module, not utils for the tests!
"""

from unittest.mock import patch

from fabric.util import get_local_user


# Basically implementation tests, because it's not feasible to do a "real" test
# on random platforms (where we have no idea what the actual invoking user is)
class get_local_user_:
    @patch("getpass.getuser")
    def defaults_to_getpass_getuser(self, getuser):
        "defaults to getpass.getuser"
        get_local_user()
        getuser.assert_called_once_with()

    @patch("getpass.getuser", side_effect=KeyError)
    def KeyError_means_SaaS_and_thus_None(self, getuser):
        assert get_local_user() is None

    # TODO: test for ImportError+win32 once appveyor is set up as w/ invoke



import os
import time

from io import StringIO

from invoke import pty_size, CommandTimedOut
from invocations.environment import in_ci
from pytest import skip, raises
from pytest_relaxed import trap

from fabric import Connection, Config


# TODO: use pytest markers
def skip_outside_ci():
    if not in_ci():
        skip()


class Connection_:
    class ssh_connections:
        def open_method_generates_real_connection(self):
            c = Connection("localhost")
            c.open()
            assert c.client.get_transport().active is True
            assert c.is_connected is True
            return c

        def close_method_closes_connection(self):
            # Handy shortcut - open things up, then return Connection for us to
            # close
            c = self.open_method_generates_real_connection()
            c.close()
            assert c.client.get_transport() is None
            assert c.is_connected is False

    class run:
        def simple_command_on_host(self):
            """
            Run command on localhost
            """
            result = Connection("localhost").run("echo foo", hide=True)
            assert result.stdout == "foo\n"
            assert result.exited == 0
            assert result.ok is True

        def simple_command_with_pty(self):
            """
            Run command under PTY on localhost
            """
            # Most Unix systems should have stty, which asplodes when not run
            # under a pty, and prints useful info otherwise
            result = Connection("localhost").run(
                "stty size", hide=True, pty=True
            )
            found = result.stdout.strip().split()
            cols, rows = pty_size()
            assert tuple(map(int, found)), rows == cols
            # PTYs use \r\n, not \n, line separation
            assert "\r\n" in result.stdout
            assert result.pty is True

    class shell:
        @trap
        def base_case(self):
            result = Connection("localhost").shell(
                # Some extra newlines to make sure it doesn't get split up by
                # motd/prompt
                in_stream=StringIO("\n\nexit\n")
            )
            assert result.command is None
            assert "exit" in result.stdout
            assert result.stderr == ""
            assert result.exited == 0
            assert result.pty is True

    class local:
        def wraps_invoke_run(self):
            # NOTE: most of the interesting tests about this are in
            # invoke.runners / invoke.integration.
            cxn = Connection("localhost")
            result = cxn.local("echo foo", hide=True)
            assert result.stdout == "foo\n"
            assert not cxn.is_connected  # meh way of proving it didn't use SSH

    def mixed_use_of_local_and_run(self):
        """
        Run command truly locally, and over SSH via localhost
        """
        cxn = Connection("localhost")
        result = cxn.local("echo foo", hide=True)
        assert result.stdout == "foo\n"
        assert not cxn.is_connected  # meh way of proving it didn't use SSH yet
        result = cxn.run("echo foo", hide=True)
        assert cxn.is_connected  # NOW it's using SSH
        assert result.stdout == "foo\n"

    class sudo:
        def setup(self):
            # NOTE: assumes a user configured for passworded (NOT
            # passwordless)_sudo, whose password is 'mypass', is executing the
            # test suite. I.e. our travis-ci setup.
            config = Config(
                {"sudo": {"password": "mypass"}, "run": {"hide": True}}
            )
            self.cxn = Connection("localhost", config=config)

        def sudo_command(self):
            """
            Run command via sudo on host localhost
            """
            skip_outside_ci()
            assert self.cxn.sudo("whoami").stdout.strip() == "root"

        def mixed_sudo_and_normal_commands(self):
            """
            Run command via sudo, and not via sudo, on localhost
            """
            skip_outside_ci()
            logname = os.environ["LOGNAME"]
            assert self.cxn.run("whoami").stdout.strip() == logname
            assert self.cxn.sudo("whoami").stdout.strip() == "root"

    def large_remote_commands_finish_cleanly(self):
        # Guards against e.g. cleanup finishing before actually reading all
        # data from the remote end. Which is largely an issue in Invoke-level
        # code but one that only really manifests when doing stuff over the
        # network. Yay computers!
        path = "/usr/share/dict/words"
        cxn = Connection("localhost")
        with open(path) as fd:
            words = [x.strip() for x in fd.readlines()]
        stdout = cxn.run("cat {}".format(path), hide=True).stdout
        lines = [x.strip() for x in stdout.splitlines()]
        # When bug present, # lines received is significantly fewer than the
        # true count in the file (by thousands).
        assert len(lines) == len(words)

    class command_timeout:
        def setup(self):
            self.cxn = Connection("localhost")

        def does_not_raise_exception_when_under_timeout(self):
            assert self.cxn.run("sleep 1", timeout=3)

        def raises_exception_when_over_timeout(self):
            with raises(CommandTimedOut) as info:
                start = time.time()
                self.cxn.run("sleep 5", timeout=1)
            elapsed = time.time() - start
            assert info.value.timeout == 1
            # Catch scenarios where we except but don't actually shut down
            # early (w/ a bit of fudge time for overhead)
            assert elapsed <= 2



from getpass import getpass
from pathlib import Path
from unittest.mock import Mock, patch

from invoke.vendor.lexicon import Lexicon
from pytest import raises, fixture
from paramiko import (
    AgentKey,
    AuthFailure,
    InMemoryPrivateKey,
    OnDiskPrivateKey,
    PKey,
    Password,
    SSHConfig,
)

from fabric import Config, OpenSSHAuthStrategy


@fixture(autouse=True)  # under NO circumstances do we wanna talk to an agent
def fake_agent():
    with patch("fabric.auth.Agent") as Agent:
        yield Agent


@fixture
def fake(fake_agent):
    """
    Bag-o-fakes:
    - Agent (do NOT wanna talk to a real one)
    - PKey (do NOT want to load real ones from disk)
    """
    lex = Lexicon()
    lex.Agent = fake_agent
    with patch("fabric.auth.PKey") as PKey:
        lex.PKey = PKey
        yield lex


def _strategy(py_keys=None, ssh_keys=None):
    conf = SSHConfig().lookup("host")
    conf["identityfile"] = ssh_keys or []
    return OpenSSHAuthStrategy(
        ssh_config=conf,
        fabric_config=Config(
            overrides={"authentication": {"identities": py_keys or []}}
        ),
        username="whatever",
    )


class OpenSSHAuthStrategy_:
    def init_accepts_additional_args_and_creates_agent(self, fake):
        with raises(TypeError):
            # Just ssh_config == not good enough
            OpenSSHAuthStrategy(None)
        with raises(TypeError):
            OpenSSHAuthStrategy(None, None)
        ssh_config, fabric_config, username = object(), object(), "foo"
        strat = OpenSSHAuthStrategy(ssh_config, fabric_config, username)
        assert strat.ssh_config is ssh_config
        assert strat.config is fabric_config
        assert strat.username is username
        fake.Agent.assert_called_once_with()
        assert strat.agent is fake.Agent.return_value

    @patch("fabric.auth.partial")
    def get_sources_yields_in_specific_order(self, mock_partial):
        # Yields from get_pubkeys
        strat = _strategy()
        strat.get_pubkeys = Mock(return_value=iter([1, 2, 3]))
        generator = strat.get_sources()
        source = next(generator)
        strat.get_pubkeys.assert_called_once_with()
        assert source == 1
        assert next(generator) == 2
        assert next(generator) == 3
        # Then tries password auth
        pw = next(generator)
        assert isinstance(pw, Password)
        assert pw.username == "whatever"
        getter = pw.password_getter
        assert getter == mock_partial.return_value
        mock_partial.assert_called_once_with(getpass, "whatever's password: ")
        # Safety check
        with raises(StopIteration):
            next(generator)
        # TODO/NOTE: this is, obviously, incomplete!

    def authenticate_always_closes_agent(self):
        strat = _strategy()
        strat.close = Mock()
        oops = Exception("onoz")
        kaboom = Mock(authenticate=Mock(side_effect=oops))
        strat.get_sources = Mock(return_value=iter([kaboom]))
        with raises(AuthFailure):
            strat.authenticate(None)
        strat.close.assert_called_once_with()

    def close_closes_agent(self):
        agent = Mock()
        strat = _strategy()
        strat.agent = agent
        strat.close()
        agent.close.assert_called_once_with()

    class get_pubkeys:
        class fabric_config:
            def loads_identities_config_var(self, fake):
                strat = _strategy(py_keys=["rsa.key"])
                keys = list(strat.get_pubkeys())
                assert len(keys) == 1
                key = keys[0]
                assert isinstance(key, OnDiskPrivateKey)
                fake.PKey.from_path.assert_called_once_with("rsa.key")
                assert key.pkey is fake.PKey.from_path.return_value
                assert key.source == "python-config"
                assert key.path == "rsa.key"

            def silently_skips_nonexistent_files(self, fake):
                key = Mock()
                fake.PKey.from_path.side_effect = [FileNotFoundError, key]
                strat = _strategy(py_keys=["rsa.key", "ed25519.key"])
                keys = list(strat.get_pubkeys())
                assert len(keys) == 1
                assert keys[0].pkey is key
                assert keys[0].path == "ed25519.key"

        class ssh_config:
            def loads_identityfile_key(self, fake):
                strat = _strategy(ssh_keys=["rsa.key"])
                keys = list(strat.get_pubkeys())
                assert len(keys) == 1
                key = keys[0]
                assert isinstance(key, OnDiskPrivateKey)
                fake.PKey.from_path.assert_called_once_with("rsa.key")
                assert key.pkey is fake.PKey.from_path.return_value
                assert key.source == "ssh-config"
                assert key.path == "rsa.key"

            def silently_skips_nonexistent_files(self, fake):
                key = Mock()
                fake.PKey.from_path.side_effect = [FileNotFoundError, key]
                strat = _strategy(ssh_keys=["rsa.key", "ed25519.key"])
                keys = list(strat.get_pubkeys())
                assert len(keys) == 1
                assert keys[0].pkey is key
                assert keys[0].path == "ed25519.key"

        class implicit_user_home_locations:
            def loads_all_four_known_key_types(self, fake):
                strat = _strategy()
                keys = list(strat.get_pubkeys())
                assert [x.path for x in keys] == [
                    Path.home() / ".ssh" / key
                    for key in [
                        "id_rsa",
                        "id_ecdsa",
                        "id_ed25519",
                        "id_dsa",
                    ]
                ]
                assert all(x.source == "implicit-home" for x in keys)

            def silently_skips_nonexistent_files(self, fake):
                fake.PKey.from_path.side_effect = [
                    FileNotFoundError,
                    Mock(),
                    Mock(),
                    Mock(),
                ]
                strat = _strategy()
                keys = list(strat.get_pubkeys())
                assert [x.path for x in keys] == [
                    Path.home() / ".ssh" / key
                    for key in [
                        # "id_rsa",  # not found
                        "id_ecdsa",
                        "id_ed25519",
                        "id_dsa",
                    ]
                ]

            def does_not_load_if_config_based_keys_given(self, fake):
                strat = _strategy(py_keys=["rsa.key"])
                keys = list(strat.get_pubkeys())
                # NOTE: no $HOME keys were loaded by PKey.from_path.
                assert len(keys) == 1
                fake.PKey.from_path.assert_called_once_with("rsa.key")

            @patch("fabric.auth.win32", True)
            def uses_windows_style_ssh_dir_on_windows(self, fake):
                strat = _strategy()
                keys = list(strat.get_pubkeys())
                assert [x.path for x in keys] == [
                    # NOTE: not .ssh, but just ssh
                    Path.home() / "ssh" / key
                    for key in [
                        "id_rsa",
                        "id_ecdsa",
                        "id_ed25519",
                        "id_dsa",
                    ]
                ]

        def loads_keys_from_agent(self, fake):
            # No $HOME keys to gum things up.
            fake.PKey.from_path.side_effect = FileNotFoundError
            agent_keys = [
                AgentKey(fake.Agent.return_value, x)
                for x in (b"dummy", b"data")
            ]
            get_keys = fake.Agent.return_value.get_keys
            get_keys.return_value = agent_keys
            strat = _strategy()
            keys = list(strat.get_pubkeys())
            get_keys.assert_called_once_with()
            assert all(isinstance(x, InMemoryPrivateKey) for x in keys)
            assert [x.pkey for x in keys] == agent_keys

        def yields_sources_in_specific_order(self, fake):
            # Set up fake-enough keys
            # Reminder: 'CLI' in our world generally means
            # CLI-or-Fabric-Config; we inject them via Fabric config.
            class FaKey(PKey):
                def __init__(self, name):
                    self._name = name
                    self.public_blob = None
                    if name.endswith(".cert"):
                        self.public_blob = True  # good enough for AuthStrategy

                @property
                def _fields(self):
                    return [self._name, self.public_blob]

                @property
                def fingerprint(self):
                    return self._name

            # Both a useful lookup source for the fake constructor, and the
            # final expected order of things.
            expected_keys = [
                # Certs, from ssh conf files
                FaKey("ssh-conf.cert"),
                # Certs, from CLI / Python conf files
                FaKey("py-conf.cert"),
                # Agent keys, when overlapping with config file keys
                FaKey("agent-conf.key"),
                # Agent keys, the rest
                FaKey("agent-noconf.key"),
                # Non-cert keys, from the CLI
                FaKey("py-conf.key"),
                # Non-cert keys, from SSH configs
                FaKey("ssh-conf.key"),
            ]

            def get_key(name):
                for candidate in expected_keys:
                    if candidate._name == name:
                        return candidate
                raise Exception(f"Your candidate list has no {name!r}!")

            fake.PKey.from_path.side_effect = get_key
            fake.Agent.return_value.get_keys.return_value = [
                # not also found in config
                get_key("agent-noconf.key"),
                # also found in config, and should show up before agent_noconf
                # does, despite showing up in get_keys() later!
                get_key("agent-conf.key"),
            ]
            strat = _strategy(
                py_keys=["py-conf.key", "py-conf.cert"],
                ssh_keys=["ssh-conf.key", "ssh-conf.cert", "agent-conf.key"],
            )
            keys = list(strat.get_pubkeys())
            # Order check!
            assert [x.pkey for x in keys] == expected_keys
